import math
import os
import warnings
import zipfile
from pathlib import Path
from typing import Any, Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from metatensor.learn.data import Dataset, group_and_join
from metatensor.learn.data._namedtuple import namedtuple
from metatensor.torch import TensorMap, load_buffer
from metatensor.torch import save_buffer as mts_save_buffer
from metatensor.torch.atomistic import System, load_system
from metatensor.torch.atomistic import save as mta_save
from omegaconf import DictConfig
from torch.utils.data import Subset

from ..external_naming import to_external_name
from ..units import get_gradient_units
from .readers.metatensor import _check_tensor_map_metadata, _empty_tensor_map_like
from .target_info import TargetInfo, get_energy_target_info, get_generic_target_info


class DatasetInfo:
    """A class that contains information about datasets.

    This class is used to communicate additional dataset details to the
    training functions of the individual models.

    :param length_unit: Unit of length used in the dataset. Examples are ``"angstrom"``
        or ``"nanometer"``.
    :param atomic_types: List containing all integer atomic types present in the
        dataset. ``atomic_types`` will be stored as a sorted list of **unique** atomic
        types.
    :param targets: Information about targets in the dataset.
    """

    def __init__(
        self, length_unit: str, atomic_types: List[int], targets: Dict[str, TargetInfo]
    ):
        self.length_unit = length_unit if length_unit is not None else ""
        self._atomic_types = set(atomic_types)
        self.targets = targets

    @property
    def atomic_types(self) -> List[int]:
        """Sorted list of unique integer atomic types."""
        return sorted(self._atomic_types)

    @atomic_types.setter
    def atomic_types(self, value: List[int]):
        self._atomic_types = set(value)

    def __repr__(self):
        return (
            f"DatasetInfo(length_unit={self.length_unit!r}, "
            f"atomic_types={self.atomic_types!r}, targets={self.targets!r})"
        )

    def __eq__(self, other):
        if not isinstance(other, DatasetInfo):
            raise NotImplementedError(
                "Comparison between a DatasetInfo instance and a "
                f"{type(other).__name__} instance is not implemented."
            )
        return (
            self.length_unit == other.length_unit
            and self._atomic_types == other._atomic_types
            and self.targets == other.targets
        )

    def copy(self) -> "DatasetInfo":
        """Return a shallow copy of the DatasetInfo."""
        return DatasetInfo(
            length_unit=self.length_unit,
            atomic_types=self.atomic_types.copy(),
            targets=self.targets.copy(),
        )

    def update(self, other: "DatasetInfo") -> None:
        """Update this instance with the union of itself and ``other``.

        :raises ValueError: If the ``length_units`` are different.
        """
        if self.length_unit != other.length_unit:
            raise ValueError(
                "Can't update DatasetInfo with a different `length_unit`: "
                f"({self.length_unit} != {other.length_unit})"
            )

        self.atomic_types = self.atomic_types + other.atomic_types

        intersecting_target_keys = self.targets.keys() & other.targets.keys()
        for key in intersecting_target_keys:
            if not self.targets[key].is_compatible_with(other.targets[key]):
                raise ValueError(
                    f"Can't update DatasetInfo with different target information for "
                    f"target '{key}': {self.targets[key]} is not compatible with "
                    f"{other.targets[key]}. If the units, quantity and keys of the two "
                    "targets are the same, this must be due to a mismatch in the "
                    "internal metadata of the layout."
                )
        self.targets.update(other.targets)

    def union(self, other: "DatasetInfo") -> "DatasetInfo":
        """Return the union of this instance with ``other``."""
        new = self.copy()
        new.update(other)
        return new


def get_stats(dataset: Union[Dataset, Subset], dataset_info: DatasetInfo) -> str:
    """Returns the statistics of a dataset or subset as a string."""

    dataset_len = len(dataset)
    stats = f"Dataset containing {dataset_len} structures"
    if dataset_len == 0:
        return stats

    # target_names will be used to store names of the targets,
    # along with their gradients
    target_names = []
    for key, tensor_map in dataset[0]._asdict().items():
        if key == "system":
            continue
        target_names.append(key)
        gradients_list = tensor_map.block(0).gradients_list()
        for gradient in gradients_list:
            target_names.append(f"{key}_{gradient}_gradients")

    sums = {key: 0.0 for key in target_names}
    sums_of_squares = {key: 0.0 for key in target_names}
    n_elements = {key: 0 for key in target_names}
    for sample in dataset:
        for key in target_names:
            if "_gradients" not in key:  # not a gradient
                tensors = [block.values for block in sample[key].blocks()]
            else:
                original_key = key.split("_")[0]
                gradient_name = key.replace(f"{original_key}_", "").replace(
                    "_gradients", ""
                )
                tensors = [
                    block.gradient(gradient_name).values
                    for block in sample[original_key].blocks()
                ]
            sums[key] += sum(tensor.sum() for tensor in tensors)
            sums_of_squares[key] += sum((tensor**2).sum() for tensor in tensors)
            n_elements[key] += sum(tensor.numel() for tensor in tensors)
    means = {key: sums[key] / n_elements[key] for key in target_names}
    means_of_squares = {
        key: sums_of_squares[key] / n_elements[key] for key in target_names
    }
    stds = {
        key: (means_of_squares[key] - means[key] ** 2) ** 0.5 for key in target_names
    }

    # Find units
    units = {}
    for key in target_names:
        # Gets the units of an output
        if key.endswith("_gradients"):
            # handling <base_name>_<gradient_name>_gradients
            base_name = key[:-10]
            gradient_name = base_name.split("_")[-1]
            base_name = base_name.replace(f"_{gradient_name}", "")
            base_unit = dataset_info.targets[base_name].unit
            unit = get_gradient_units(
                base_unit, gradient_name, dataset_info.length_unit
            )
        else:
            unit = dataset_info.targets[key].unit
        units[key] = unit

    stats += "\n    Mean and standard deviation of targets:"
    for key in target_names:
        stats += (
            f"\n    - {to_external_name(key, dataset_info.targets)}: "  # type: ignore
            + f"\n      - mean {means[key]:.4g}"
            + (f" {units[key]}" if units[key] != "" else "")
            + f"\n      - std  {stds[key]:.4g}"
            + (f" {units[key]}" if units[key] != "" else "")
        )

    return stats


def get_atomic_types(datasets: Union[Dataset, List[Dataset]]) -> List[int]:
    """List of all atomic types present in a dataset or list of datasets.

    :param datasets: the dataset, or list of datasets
    :returns: sorted list of all atomic types present in the datasets
    """

    if not isinstance(datasets, list):
        datasets = [datasets]

    types = set()
    for dataset in datasets:
        for sample in dataset:
            system = sample["system"]
            types.update(set(system.types.tolist()))

    return sorted(types)


def get_all_targets(datasets: Union[Dataset, List[Dataset]]) -> List[str]:
    """Sorted list of all unique targets present in a dataset or list of datasets.

    :param datasets: the dataset(s).
    :returns: Sorted list of all targets present in the dataset(s).
    """

    if not isinstance(datasets, list):
        datasets = [datasets]

    # The following does not work because the `dataset` can also
    # be a `Subset` object:
    # return list(dataset.targets.keys())

    # Iterate over all single instances of the dataset:
    target_names = []
    for dataset in datasets:
        for sample in dataset:
            # system not needed
            target_names += [key for key in sample._asdict().keys() if key != "system"]

    return sorted(set(target_names))


def collate_fn(batch: List[Dict[str, Any]]) -> Tuple[List, Dict[str, TensorMap]]:
    """
    Wraps `group_and_join` to
    return the data fields as a list of systems, and a dictionary of nameed
    targets.
    """

    collated_targets = group_and_join(batch, join_kwargs={"remove_tensor_name": True})
    collated_targets = collated_targets._asdict()
    systems = collated_targets.pop("system")
    return systems, collated_targets


def check_datasets(train_datasets: List[Dataset], val_datasets: List[Dataset]):
    """Check that the training and validation sets are compatible with one another

    Although these checks will not fit all use cases, most models would be expected
    to be able to use this function.

    :param train_datasets: A list of training datasets to check.
    :param val_datasets: A list of validation datasets to check
    :raises TypeError: If the ``dtype`` within the datasets are inconsistent.
    :raises ValueError: If the `val_datasets` has a target that is not present in
        the ``train_datasets``.
    :raises ValueError: If the training or validation set contains chemical species
        or targets that are not present in the training set
    """
    # Check that system `dtypes` are consistent within datasets
    desired_dtype = None
    for train_dataset in train_datasets:
        if len(train_dataset) == 0:
            continue

        actual_dtype = train_dataset[0].system.positions.dtype
        if desired_dtype is None:
            desired_dtype = actual_dtype

        if actual_dtype != desired_dtype:
            raise TypeError(
                "`dtype` between datasets is inconsistent, "
                f"found {desired_dtype} and {actual_dtype} in training datasets"
            )

    for val_dataset in val_datasets:
        if len(val_dataset) == 0:
            continue

        actual_dtype = val_dataset[0].system.positions.dtype

        if desired_dtype is None:
            desired_dtype = actual_dtype

        if actual_dtype != desired_dtype:
            raise TypeError(
                "`dtype` between datasets is inconsistent, "
                f"found {desired_dtype} and {actual_dtype} in validation datasets"
            )

    # Get all targets in the training and validation sets:
    train_targets = get_all_targets(train_datasets)
    val_targets = get_all_targets(val_datasets)

    # Check that the validation sets do not have targets that are not in the
    # training sets:
    for target in val_targets:
        if target not in train_targets:
            raise ValueError(
                f"The validation dataset has a target ({target}) that is not present "
                "in the training dataset."
            )
    # Get all the species in the training and validation sets:
    all_train_species = get_atomic_types(train_datasets)
    all_val_species = get_atomic_types(val_datasets)

    # Check that the validation sets do not have species that are not in the
    # training sets:
    for species in all_val_species:
        if species not in all_train_species:
            raise ValueError(
                f"The validation dataset has a species ({species}) that is not in the "
                "training dataset. This could be a result of a random train/validation "
                "split. You can avoid this by providing a validation dataset manually."
            )


def _train_test_random_split(
    train_dataset: Dataset,
    train_size: float,
    test_size: float,
) -> List[Dataset]:
    if train_size <= 0:
        raise ValueError("Fraction of the train set is smaller or equal to 0!")

    # normalize the sizes
    size_sum = train_size + test_size
    train_size /= size_sum
    test_size /= size_sum

    # find number of samples in the train and test sets
    test_len = math.floor(len(train_dataset) * test_size)
    if test_len == 0:
        warnings.warn(
            "Requested dataset of zero length. This dataset will be empty.",
            UserWarning,
            stacklevel=2,
        )
    train_len = len(train_dataset) - test_len
    if train_len == 0:
        raise ValueError("No samples left in the training set.")

    # find train, test indices
    indices = list(range(len(train_dataset)))
    np.random.shuffle(indices)
    train_indices = indices[:train_len]
    test_indices = indices[train_len:]

    return [
        Subset(train_dataset, train_indices),
        Subset(train_dataset, test_indices),
    ]


class DiskDataset(torch.utils.data.Dataset):
    """A class representing a dataset stored on disk.

    The dataset is stored in a zip file, where each sample is stored in a separate
    directory. The directory's name is the index of the sample (e.g. ``0/``), and the
    files in the directory are the system (``system.mta``) and the targets
    (each named ``<target_name>.mts``). These are ``metatensor.torch.atomistic.System``
    and ``metatensor.torch.TensorMap`` objects, respectively.

    Such a dataset can be created conveniently using the :py:class:`DiskDatasetWriter`
    class.

    :param path: Path to the zip file containing the dataset.
    """

    def __init__(self, path: Union[str, Path]):
        self.zip_file = zipfile.ZipFile(path, "r")
        self._field_names = ["system"]
        # check that we have at least one sample:
        if "0/system.mta" not in self.zip_file.namelist():
            raise ValueError(
                "Could not find `0/system.mta` in the zip file. "
                "The dataset format might be wrong, or the dataset might be empty. "
                "Empty disk datasets are not supported."
            )
        for file_name in self.zip_file.namelist():
            if file_name.startswith("0/") and file_name.endswith(".mts"):
                self._field_names.append(file_name[2:-4])
        self._sample_class = namedtuple("Sample", self._field_names)
        self._len = len([f for f in self.zip_file.namelist() if f.endswith(".mta")])

    def __len__(self):
        return self._len

    def __getitem__(self, index):
        system_and_targets = []
        for field_name in self._field_names:
            if field_name == "system":
                with self.zip_file.open(f"{index}/system.mta", "r") as file:
                    system = load_system(file)
                    system_and_targets.append(system)
            else:
                with self.zip_file.open(f"{index}/{field_name}.mts", "r") as file:
                    numpy_buffer = np.load(file)
                    tensor_buffer = torch.from_numpy(numpy_buffer)
                    tensor_map = load_buffer(tensor_buffer)
                    system_and_targets.append(tensor_map)
        return self._sample_class(*system_and_targets)

    def __del__(self):
        self.zip_file.close()

    def get_target_info(self, target_config: DictConfig) -> Dict[str, TargetInfo]:
        """
        Get information about the targets in the dataset.

        :param target_config: The user-provided (through the yaml file) target
            configuration.
        """
        target_info_dict = {}
        for target_key, target in target_config.items():
            is_energy = (
                (target["quantity"] == "energy")
                and (not target["per_atom"])
                and target["num_subtargets"] == 1
                and target["type"] == "scalar"
            )
            tensor_map = self[0][target_key]  # always > 0 samples, see above
            if is_energy:
                if len(tensor_map) != 1:
                    raise ValueError("Energy TensorMaps should have exactly one block.")
                add_position_gradients = tensor_map.block().has_gradient("positions")
                add_strain_gradients = tensor_map.block().has_gradient("strain")
                target_info = get_energy_target_info(
                    target, add_position_gradients, add_strain_gradients
                )
                _check_tensor_map_metadata(tensor_map, target_info.layout)
                target_info_dict[target_key] = target_info
            else:
                target_info = get_generic_target_info(target)
                _check_tensor_map_metadata(tensor_map, target_info.layout)
                # make sure that the properties of the target_info.layout also match the
                # actual properties of the tensor maps
                target_info.layout = _empty_tensor_map_like(tensor_map)
                target_info_dict[target_key] = target_info
        return target_info_dict


def _is_disk_dataset(dataset: Any) -> bool:
    # this also needs to detect if it's a ``torch.nn.utils.data.Subset`` object
    # with a ``DiskDataset`` object as its dataset, recursively
    if isinstance(dataset, DiskDataset):
        return True
    if isinstance(dataset, torch.utils.data.Subset):
        return _is_disk_dataset(dataset.dataset)
    return False


class DiskDatasetWriter:
    """
    A class for writing a dataset to disk, to be read by the :py:class:`DiskDataset`
    class.

    The class is initialized with a path to a zip file, and samples can be written to
    the zip file using the :py:meth:`write_sample` method.

    :param path: Path to the zip file to write the dataset to.
    """

    def __init__(self, path: Union[str, Path]):
        self.zip_file = zipfile.ZipFile(path, "w")
        self.index = 0

    def write_sample(self, system: System, targets: Dict[str, TensorMap]):
        """
        Write a sample to the zip file.

        :param system: The system to write.
        :param targets: A dictionary of targets to write, where each value is
            a :py:class:`TensorMap`.
        """
        with self.zip_file.open(f"{self.index}/system.mta", "w") as file:
            mta_save(file, system)
        for target_name, target in targets.items():
            with self.zip_file.open(f"{self.index}/{target_name}.mts", "w") as file:
                tensor_buffer = mts_save_buffer(target)
                numpy_buffer = tensor_buffer.numpy()
                np.save(file, numpy_buffer)
        self.index += 1

    def __del__(self):
        self.zip_file.close()


def _save_indices(
    train_indices: List[Optional[List[int]]],
    val_indices: List[Optional[List[int]]],
    test_indices: List[Optional[List[int]]],
    checkpoint_dir: Union[str, Path],
) -> None:
    # Save the indices of the training, validation, and test sets to the checkpoint
    # directory. This is useful for plotting errors and similar.

    # case 1: all indices are None (i.e. all datasets were user-provided explicitly)
    if all(indices is None for indices in train_indices):
        pass

    # case 2: there is only one dataset
    elif len(train_indices) == 1:  # val and test are the same length
        os.mkdir(os.path.join(checkpoint_dir, "indices/"))
        if train_indices[0] is not None:
            np.savetxt(
                os.path.join(checkpoint_dir, "indices/training.txt"),
                train_indices[0],
                fmt="%d",
            )
        if val_indices[0] is not None:
            np.savetxt(
                os.path.join(checkpoint_dir, "indices/validation.txt"),
                val_indices[0],
                fmt="%d",
            )
        if test_indices[0] is not None:
            np.savetxt(
                os.path.join(checkpoint_dir, "indices/test.txt"),
                test_indices[0],
                fmt="%d",
            )

    # case 3: there are multiple datasets
    else:
        os.mkdir(os.path.join(checkpoint_dir, "indices/"))
        for i, (train, val, test) in enumerate(
            zip(train_indices, val_indices, test_indices)
        ):
            if train is not None:
                np.savetxt(
                    os.path.join(checkpoint_dir, f"indices/training_{i}.txt"),
                    train,
                    fmt="%d",
                )
            if val is not None:
                np.savetxt(
                    os.path.join(checkpoint_dir, f"indices/validation_{i}.txt"),
                    val,
                    fmt="%d",
                )
            if test is not None:
                np.savetxt(
                    os.path.join(checkpoint_dir, f"indices/test_{i}.txt"),
                    test,
                    fmt="%d",
                )



import ase.io
import numpy as np
import torch
from matscipy.neighbours import neighbour_list as neighbor_list
from torch_geometric.data import Data

from .long_range import get_all_k, get_reciprocal, get_volume


class Molecule:
    def __init__(
        self,
        atoms,
        r_cut,
        use_additional_scalar_attributes,
        use_long_range,
        k_cut,
        multi_target,
        target_index_key,
    ):
        self.use_additional_scalar_attributes = use_additional_scalar_attributes

        self.atoms = atoms

        positions = self.atoms.get_positions()
        species = self.atoms.get_atomic_numbers()

        self.central_species = []
        for i in range(len(positions)):
            self.central_species.append(species[i])

        if use_additional_scalar_attributes:
            scalar_attributes = self.atoms.arrays["scalar_attributes"]
            if len(scalar_attributes.shape) == 1:
                scalar_attributes = scalar_attributes[:, np.newaxis]

            self.central_scalar_attributes = scalar_attributes

        i_list, j_list, D_list, S_list = ase.neighborlist.neighbor_list(
            "ijDS", atoms, r_cut
        )

        self.neighbors_index = [[] for i in range(len(positions))]
        self.neighbors_shift = [[] for i in range(len(positions))]

        for i, j, _D, S in zip(i_list, j_list, D_list, S_list):
            self.neighbors_index[i].append(j)
            self.neighbors_shift[i].append(S)

        self.relative_positions = [[] for i in range(len(positions))]
        self.neighbor_species = [[] for i in range(len(positions))]
        self.neighbors_pos = [[] for i in range(len(positions))]

        if use_additional_scalar_attributes:
            self.neighbor_scalar_attributes = [[] for i in range(len(positions))]

        def is_same(first, second):
            for i in range(len(first)):
                if first[i] != second[i]:
                    return False
            return True

        for i, j, D, S in zip(i_list, j_list, D_list, S_list):
            self.relative_positions[i].append(D)
            self.neighbor_species[i].append(species[j])
            if use_additional_scalar_attributes:
                self.neighbor_scalar_attributes[i].append(scalar_attributes[j])
            for k in range(len(self.neighbors_index[j])):
                if (self.neighbors_index[j][k] == i) and is_same(
                    self.neighbors_shift[j][k], -S
                ):
                    self.neighbors_pos[i].append(k)

        self.use_long_range = use_long_range
        if self.use_long_range:
            self.cell = np.array(self.atoms.get_cell())
            w_1, w_2, w_3 = get_reciprocal(self.cell[0], self.cell[1], self.cell[2])
            reciprocal = np.concatenate(
                [w_1[np.newaxis], w_2[np.newaxis], w_3[np.newaxis]], axis=0
            )
            self.reciprocal = reciprocal
            self.k_vectors = get_all_k(self.cell[0], self.cell[1], self.cell[2], k_cut)
            self.k_cut = k_cut
            self.volume = get_volume(self.cell[0], self.cell[1], self.cell[2])

        if multi_target:
            self.target_index = int(self.atoms.info[target_index_key])
        else:
            self.target_index = None

    def get_max_num(self):
        maximum = None
        for chunk in self.relative_positions:
            if (maximum is None) or (len(chunk) > maximum):
                maximum = len(chunk)
        return maximum

    def get_num_k(self):
        if self.use_long_range:
            return len(self.k_vectors)
        else:
            return None

    def get_graph(self, max_num, all_species, max_num_k):
        central_species = [
            np.where(all_species == specie)[0][0] for specie in self.central_species
        ]
        central_species = torch.LongTensor(central_species)

        nums = []
        mask = []
        relative_positions = np.zeros([len(self.relative_positions), max_num, 3])
        neighbors_pos = np.zeros([len(self.relative_positions), max_num], dtype=int)
        neighbors_index = np.zeros([len(self.relative_positions), max_num], dtype=int)

        if self.use_additional_scalar_attributes:
            neighbor_scalar_attributes = np.zeros(
                [len(self.relative_positions), max_num, 1]
            )

        for i in range(len(self.relative_positions)):
            now = np.array(self.relative_positions[i])
            if len(now) > 0:
                if self.use_additional_scalar_attributes:
                    neighbor_scalar_attributes[i, : len(now)] = (
                        self.neighbor_scalar_attributes[i]
                    )
                relative_positions[i, : len(now), :] = now
                neighbors_pos[i, : len(now)] = self.neighbors_pos[i]
                neighbors_index[i, : len(now)] = self.neighbors_index[i]

            nums.append(len(self.relative_positions[i]))
            current_mask = np.zeros(max_num)
            current_mask[len(self.relative_positions[i]) :] = True
            mask.append(current_mask[np.newaxis, :])

        mask = np.concatenate(mask, axis=0)
        relative_positions = torch.tensor(
            relative_positions, dtype=torch.get_default_dtype()
        )
        nums = torch.tensor(nums, dtype=torch.get_default_dtype())
        mask = torch.BoolTensor(mask)

        neighbors_pos = torch.LongTensor(neighbors_pos)
        neighbors_index = torch.LongTensor(neighbors_index)

        neighbor_species = len(all_species) * np.ones(
            [len(self.neighbor_species), max_num], dtype=int
        )
        for i in range(len(self.neighbor_species)):
            now = np.array(self.neighbor_species[i])
            now = np.array([np.where(all_species == specie)[0][0] for specie in now])
            neighbor_species[i, : len(now)] = now
        neighbor_species = torch.LongTensor(neighbor_species)

        kwargs = {
            "central_species": central_species,
            "x": relative_positions,
            "neighbor_species": neighbor_species,
            "neighbors_pos": neighbors_pos,
            "neighbors_index": neighbors_index.transpose(0, 1),
            "nums": nums,
            "mask": mask,
            "n_atoms": len(self.atoms.positions),
        }

        if self.target_index is not None:
            kwargs["target_id"] = self.target_index

        if self.use_additional_scalar_attributes:
            kwargs["neighbor_scalar_attributes"] = torch.tensor(
                neighbor_scalar_attributes, dtype=torch.get_default_dtype()
            )
            kwargs["central_scalar_attributes"] = torch.tensor(
                self.central_scalar_attributes, dtype=torch.get_default_dtype()
            )

        if self.use_long_range:
            kwargs["cell"] = torch.tensor(self.cell, dtype=torch.get_default_dtype())[
                None
            ]
            kwargs["reciprocal"] = torch.tensor(
                self.reciprocal, dtype=torch.get_default_dtype()
            )[None]
            k_vectors = np.zeros([1, max_num_k, 3])
            k_mask = np.zeros([max_num_k], dtype=bool)
            for index in range(len(self.k_vectors)):
                k_vectors[0, index] = self.k_vectors[index]
                k_mask[index] = True
            kwargs["k_vectors"] = torch.tensor(
                k_vectors, dtype=torch.get_default_dtype()
            )
            kwargs["k_mask"] = torch.BoolTensor(k_mask)[None]

            kwargs["positions"] = torch.tensor(
                self.atoms.get_positions(), dtype=torch.get_default_dtype()
            )
            kwargs["volume"] = torch.tensor(
                self.volume, dtype=torch.get_default_dtype()
            )

        result = Data(**kwargs)

        return result


class MoleculeCPP:
    def __init__(
        self,
        atoms,
        r_cut,
        use_additional_scalar_attributes,
        use_long_range,
        k_cut,
        multi_target,
        target_index_key,
    ):
        self.use_additional_scalar_attributes = use_additional_scalar_attributes
        self.atoms = atoms
        self.r_cut = r_cut
        self.use_long_range = use_long_range
        self.k_cut = k_cut

        if self.use_long_range:
            raise NotImplementedError("Long range is not implemented in cpp")
        if self.use_additional_scalar_attributes:
            raise NotImplementedError(
                "Additional scalar attributes are not implemented in cpp"
            )

        def is_3d_crystal(atoms):
            pbc = atoms.get_pbc()
            if isinstance(pbc, bool):
                return pbc
            return all(pbc)

        if is_3d_crystal(atoms):
            i_list, j_list, D_list, S_list = neighbor_list("ijDS", atoms, r_cut)
        else:
            i_list, j_list, D_list, S_list = ase.neighborlist.neighbor_list(
                "ijDS", atoms, r_cut
            )

        self.i_list = torch.tensor(i_list, dtype=torch.int64).contiguous()
        self.j_list = torch.tensor(j_list, dtype=torch.int64).contiguous()
        self.D_list = torch.tensor(D_list, dtype=torch.get_default_dtype()).contiguous()
        self.S_list = torch.tensor(S_list, dtype=torch.int64).contiguous()
        self.species = torch.tensor(
            atoms.get_atomic_numbers(), dtype=torch.int64
        ).contiguous()
        if len(self.i_list) == 0:
            self.max_num = 0
        else:
            self.max_num = torch.max(torch.bincount(self.i_list))

        if multi_target:
            self.target_index = int(self.atoms.info[target_index_key])
        else:
            self.target_index = None

    def get_num_k(self):
        raise NotImplementedError("Long range is not implemented in cpp")

    def get_max_num(self):
        return self.max_num

    def get_graph(self, max_num, all_species, max_num_k):
        n_atoms = len(self.atoms.get_atomic_numbers())
        all_species = torch.tensor(all_species, dtype=torch.int64).contiguous()

        (
            neighbors_index,
            relative_positions,
            nums,
            mask,
            neighbor_species,
            neighbors_pos,
            species_mapped,
        ) = torch.ops.neighbors_convert.process(
            self.i_list,
            self.j_list,
            self.S_list,
            self.D_list,
            max_num,
            n_atoms,
            self.species,
            all_species,
        )

        kwargs = {
            "central_species": species_mapped,
            "x": relative_positions,
            "neighbor_species": neighbor_species,
            "neighbors_pos": neighbors_pos,
            "neighbors_index": neighbors_index.transpose(0, 1),
            "nums": nums,
            "mask": mask,
            "n_atoms": len(self.atoms.positions),
        }

        if self.target_index is not None:
            kwargs["target_id"] = self.target_index

        result = Data(**kwargs)

        return result


def batch_to_dict(batch):
    batch_dict = {
        "x": batch.x,
        "central_species": batch.central_species,
        "neighbor_species": batch.neighbor_species,
        "mask": batch.mask,
        "batch": batch.batch,
        "nums": batch.nums,
        "neighbors_index": batch.neighbors_index.transpose(0, 1),
        "neighbors_pos": batch.neighbors_pos,
    }
    if hasattr(batch, "target_id"):
        batch_dict["target_id"] = batch.target_id

    if hasattr(batch, "neighbor_scalar_attributes"):
        batch_dict["neighbor_scalar_attributes"] = batch.neighbor_scalar_attributes
    if hasattr(batch, "central_scalar_attributes"):
        batch_dict["central_scalar_attributes"] = batch.central_scalar_attributes

    if hasattr(batch, "k_vectors"):
        batch_dict["k_vectors"] = batch.k_vectors

    if hasattr(batch, "k_mask"):
        batch_dict["k_mask"] = batch.k_mask

    if hasattr(batch, "positions"):
        batch_dict["positions"] = batch.positions

    if hasattr(batch, "volume"):
        batch_dict["volume"] = batch.volume

    return batch_dict


class NeighborIndexConstructor:
    def __init__(self, i_list, j_list, S_list, species):
        n_atoms = len(species)
        self.neighbors_index = [[] for i in range(n_atoms)]
        self.neighbors_shift = [[] for i in range(n_atoms)]

        for i, j, _index, S in zip(i_list, j_list, range(len(i_list)), S_list):
            self.neighbors_index[i].append(j)
            self.neighbors_shift[i].append(S)

        self.relative_positions = [[] for i in range(n_atoms)]
        self.neighbor_species = [[] for i in range(n_atoms)]
        self.neighbors_pos = [[] for i in range(n_atoms)]

        def is_same(first, second):
            for i in range(len(first)):
                if first[i] != second[i]:
                    return False
            return True

        for i, j, index, S in zip(i_list, j_list, range(len(i_list)), S_list):
            self.relative_positions[i].append(index)
            self.neighbor_species[i].append(species[j])
            for k in range(len(self.neighbors_index[j])):
                if (self.neighbors_index[j][k] == i) and is_same(
                    self.neighbors_shift[j][k], -S
                ):
                    self.neighbors_pos[i].append(k)

    def get_max_num(self):
        maximum = None
        for chunk in self.relative_positions:
            if (maximum is None) or (len(chunk) > maximum):
                maximum = len(chunk)
        return maximum

    def get_neighbor_index(self, max_num, all_species):
        nums = []
        mask = []
        relative_positions = np.zeros([len(self.relative_positions), max_num])
        neighbors_pos = np.zeros([len(self.relative_positions), max_num], dtype=int)
        neighbors_index = np.zeros([len(self.relative_positions), max_num], dtype=int)

        for i in range(len(self.relative_positions)):
            now = np.array(self.relative_positions[i])
            if len(now) > 0:
                relative_positions[i, : len(now)] = now
                neighbors_pos[i, : len(now)] = self.neighbors_pos[i]
                neighbors_index[i, : len(now)] = self.neighbors_index[i]

            nums.append(len(self.relative_positions[i]))
            current_mask = np.zeros(max_num)
            current_mask[len(self.relative_positions[i]) :] = True
            mask.append(current_mask[np.newaxis, :])

        mask = np.concatenate(mask, axis=0)
        relative_positions = torch.LongTensor(relative_positions)
        nums = torch.tensor(nums, dtype=torch.get_default_dtype())
        mask = torch.BoolTensor(mask)

        neighbors_pos = torch.LongTensor(neighbors_pos)
        neighbors_index = torch.LongTensor(neighbors_index)

        neighbor_species = len(all_species) * np.ones(
            [len(self.neighbor_species), max_num], dtype=int
        )
        for i in range(len(self.neighbor_species)):
            now = np.array(self.neighbor_species[i])
            now = np.array([np.where(all_species == specie)[0][0] for specie in now])
            neighbor_species[i, : len(now)] = now
        neighbor_species = torch.LongTensor(neighbor_species)

        return (
            neighbors_pos,
            neighbors_index,
            nums,
            mask,
            neighbor_species,
            relative_positions,
        )



import numpy as np
import spherical
import torch
from metatensor.torch.atomistic import System
from scipy.spatial.transform import Rotation


def get_random_rotation() -> Rotation:
    return Rotation.random()


def rotate_system(system: System, rotation: Rotation):
    rotated_positions = (
        system.positions.detach().cpu().numpy() @ np.array(rotation.as_matrix()).T
    )
    rotated_cell = system.cell.detach().cpu().numpy() @ np.array(rotation.as_matrix()).T
    return System(
        positions=torch.tensor(
            rotated_positions, device=system.device, dtype=system.positions.dtype
        ),
        cell=torch.tensor(rotated_cell, device=system.device, dtype=system.cell.dtype),
        types=system.types,
        pbc=system.pbc,
    )


def rotate_spherical_tensor(spherical_tensor: np.ndarray, rotation: Rotation):
    # the spherical tensor is a tensor of shape (n_samples, 2*l+1, n_properties)
    L = (spherical_tensor.shape[1] - 1) // 2
    rotated_spherical_tensor = (
        spherical_tensor.swapaxes(-1, -2) @ calculate_wigner_D(rotation, L).T
    ).swapaxes(-1, -2)
    return rotated_spherical_tensor


def calculate_wigner_D(rotation, L):
    # We initialize the Wigner calculator from the quaternionic library...
    wigner = spherical.Wigner(L)
    # ...and we also initialize the transformation matrix from complex to real
    complex_to_real_transform = complex_to_real_spherical_harmonics_transform(L)

    # Obtaining the quaternion associated with the rotation by means of the scipy
    # libtary
    quaternion_scipy = rotation.as_quat()
    # Change convention to be consistent with the one of the quaternionic library
    quaternion_quaternionic = scipy_quaternion_to_quaternionic(quaternion_scipy)
    # applying the quaternion to the Wigner D to obtain the actual values
    wigners_R = wigner.D(quaternion_quaternionic)

    # We now extract the values of the Wigner D matrices and transform them to real
    wigner_D_matrix_complex = np.zeros((2 * L + 1, 2 * L + 1), dtype=np.complex128)
    for m in range(-L, L + 1):
        for mp in range(-L, L + 1):
            # This is the procedure that gives the correct indexing of the Wigner
            # D matrices, note that the quaternionic library uses a convention such
            # that the resulting matrix is the complex conjugate of the one that we
            # expect from the rotation, and so we take the complex conjugation
            wigner_D_matrix_complex[m + L, mp + L] = (
                wigners_R[wigner.Dindex(L, m, mp)]
            ).conj()

    # We finally transform everything in the real representation...
    wigner_D_matrix = (
        complex_to_real_transform.conj()
        @ wigner_D_matrix_complex
        @ complex_to_real_transform.T
    )
    # ...and check that we do not have imaginary contributions
    assert np.allclose(wigner_D_matrix.imag, 0.0)  # check that the matrix is real

    return wigner_D_matrix.real


def complex_to_real_spherical_harmonics_transform(ell: int):
    # Generates the transformation matrix from complex spherical harmonics
    # to real spherical harmonics for a given l.
    # Returns a transformation matrix of shape ((2l+1), (2l+1)).

    if ell < 0 or not isinstance(ell, int):
        raise ValueError("l must be a non-negative integer.")

    # The size of the transformation matrix is (2l+1) x (2l+1)
    size = 2 * ell + 1
    U = np.zeros((size, size), dtype=complex)

    for m in range(-ell, ell + 1):
        m_index = m + ell  # Index in the matrix
        if m > 0:
            # Real part of Y_{l}^{m}
            U[m_index, ell + m] = 1 / np.sqrt(2) * (-1) ** m
            U[m_index, ell - m] = 1 / np.sqrt(2)
        elif m < 0:
            # Imaginary part of Y_{l}^{|m|}
            U[m_index, ell + abs(m)] = -1j / np.sqrt(2) * (-1) ** m
            U[m_index, ell - abs(m)] = 1j / np.sqrt(2)
        else:  # m == 0
            # Y_{l}^{0} remains unchanged
            U[m_index, ell] = 1

    return U


def scipy_quaternion_to_quaternionic(q_scipy):
    # This function convert a quaternion obtained from the scipy library to the format
    # used by the quaternionic library.
    # Note: 'xyzw' is the format used by scipy.spatial.transform.Rotation
    # while 'wxyz' is the format used by quaternionic.
    qx, qy, qz, qw = q_scipy
    q_quaternion = np.array([qw, qx, qy, qz])
    return q_quaternion


def rotation_matrix_conversion_order(rotation_matrix):
    # This function is used to convert a rotation matrix from the format (y, z, x)
    # to (x, y, z).
    # Note: 'xyz' is the format used by scipy.spatial.transform.Rotation
    # while 'yzx' is the format used by the spherical harmonics.
    converted_matrix = rotation_matrix[[2, 0, 1], :]
    converted_matrix = converted_matrix[:, [2, 0, 1]]
    return converted_matrix



import argparse
import itertools
import logging
import time
from pathlib import Path
from typing import Dict, List, Optional, Union

import metatensor.torch
import numpy as np
import torch
from metatensor.torch import Labels, TensorBlock, TensorMap
from metatensor.torch.atomistic import MetatensorAtomisticModel
from omegaconf import DictConfig, OmegaConf

from ..utils.data import (
    Dataset,
    TargetInfo,
    collate_fn,
    read_systems,
    read_targets,
    write_predictions,
)
from ..utils.errors import ArchitectureError
from ..utils.evaluate_model import evaluate_model
from ..utils.io import load_model
from ..utils.logging import MetricLogger
from ..utils.metrics import MAEAccumulator, RMSEAccumulator
from ..utils.neighbor_lists import (
    get_requested_neighbor_lists,
    get_system_with_neighbor_lists,
)
from ..utils.omegaconf import expand_dataset_config
from ..utils.per_atom import average_by_num_atoms
from .formatter import CustomHelpFormatter


logger = logging.getLogger(__name__)


def _add_eval_model_parser(subparser: argparse._SubParsersAction) -> None:
    """Add the `eval_model` paramaters to an argparse (sub)-parser"""

    if eval_model.__doc__ is not None:
        description = eval_model.__doc__.split(r":param")[0]
    else:
        description = None

    # If you change the synopsis of these commands or add new ones adjust the completion
    # script at `src/metatrain/share/metatrain-completion.bash`.
    parser = subparser.add_parser(
        "eval",
        description=description,
        formatter_class=CustomHelpFormatter,
    )
    parser.set_defaults(callable="eval_model")
    parser.add_argument(
        "path",
        type=str,
        help="Saved exported (.pt) model to be evaluated.",
    )
    parser.add_argument(
        "options",
        type=str,
        help="Eval options YAML file to define a dataset for evaluation.",
    )
    parser.add_argument(
        "-e",
        "--extensions-dir",
        type=str,
        required=False,
        dest="extensions_directory",
        default=None,
        help=(
            "path to a directory containing all extensions required by the exported "
            "model"
        ),
    )
    parser.add_argument(
        "-o",
        "--output",
        dest="output",
        type=str,
        required=False,
        default="output.xyz",
        help="filename of the predictions (default: %(default)s)",
    )
    parser.add_argument(
        "-b",
        "--batch-size",
        dest="batch_size",
        required=False,
        type=int,
        default=1,
        help="batch size for evaluation (default: %(default)s)",
    )
    parser.add_argument(
        "--check-consistency",
        dest="check_consistency",
        action="store_true",
        help="whether to run consistency checks (default: %(default)s)",
    )


def _prepare_eval_model_args(args: argparse.Namespace) -> None:
    """Prepare arguments for eval_model."""
    args.options = OmegaConf.load(args.options)
    # models for evaluation are already exported. Don't have to pass the `name` argument
    args.model = load_model(
        path=args.__dict__.pop("path"),
        extensions_directory=args.__dict__.pop("extensions_directory"),
    )


def _concatenate_tensormaps(
    tensormap_dict_list: List[Dict[str, TensorMap]],
) -> Dict[str, TensorMap]:
    # Concatenating TensorMaps is tricky, because the model does not know the
    # "number" of the system it is predicting. For example, if a model predicts
    # 3 batches of 4 atoms each, the system labels will be [0, 1, 2, 3],
    # [0, 1, 2, 3], [0, 1, 2, 3] for the three batches, respectively. Due
    # to this, the join operation would not achieve the desired result
    # ([0, 1, 2, ..., 11, 12]). Here, we fix this by renaming the system labels.

    system_counter = 0
    n_systems = 0
    tensormaps_shifted_systems = []
    for tensormap_dict in tensormap_dict_list:
        tensormap_dict_shifted = {}
        for name, tensormap in tensormap_dict.items():
            new_keys = []
            new_blocks = []
            for key, block in tensormap.items():
                new_key = key
                where_system = block.samples.names.index("system")
                n_systems = torch.max(block.samples.column("system")) + 1
                new_samples_values = block.samples.values
                new_samples_values[:, where_system] += system_counter
                new_block = TensorBlock(
                    values=block.values,
                    samples=Labels(block.samples.names, values=new_samples_values),
                    components=block.components,
                    properties=block.properties,
                )
                for gradient_name, gradient_block in block.gradients():
                    new_block.add_gradient(
                        gradient_name,
                        gradient_block,
                    )
                new_keys.append(new_key)
                new_blocks.append(new_block)
            tensormap_dict_shifted[name] = TensorMap(
                keys=Labels(
                    names=tensormap.keys.names,
                    values=torch.stack([new_key.values for new_key in new_keys]),
                ),
                blocks=new_blocks,
            )
        tensormaps_shifted_systems.append(tensormap_dict_shifted)
        system_counter += n_systems

    return {
        target: metatensor.torch.join(
            [pred[target] for pred in tensormaps_shifted_systems], axis="samples"
        )
        for target in tensormaps_shifted_systems[0].keys()
    }


def _eval_targets(
    model: Union[MetatensorAtomisticModel, torch.jit._script.RecursiveScriptModule],
    dataset: Union[Dataset, torch.utils.data.Subset],
    options: Dict[str, TargetInfo],
    return_predictions: bool,
    batch_size: int = 1,
    check_consistency: bool = False,
) -> Optional[Dict[str, TensorMap]]:
    """Evaluates an exported model on a dataset and prints the RMSEs for each target.
    Optionally, it also returns the predictions of the model.

    The total and per-atom timings for the evaluation are also printed.

    Wraps around metatrain.cli.evaluate_model.
    """

    if len(dataset) == 0:
        logging.info("This dataset is empty. No evaluation will be performed.")
        return None

    # Attach neighbor lists to the systems:
    # TODO: these might already be present... find a way to avoid recomputing
    # if already present (e.g. if this function is called after training)
    for sample in dataset:
        system = sample["system"]
        get_system_with_neighbor_lists(system, get_requested_neighbor_lists(model))

    # Infer the device and dtype from the model
    model_tensor = next(itertools.chain(model.parameters(), model.buffers()))
    dtype = model_tensor.dtype
    device = "cpu"
    if torch.cuda.is_available() and "cuda" in model.capabilities().supported_devices:
        device = "cuda"
    logging.info(f"Running on device {device} with dtype {dtype}")
    model.to(dtype=dtype, device=device)

    if len(dataset) % batch_size != 0:
        # debug level: we don't want to clutter the output at the end of training
        # gross issues will still show up in the standard deviation of the timings
        logging.debug(
            f"The dataset size ({len(dataset)}) is not a multiple of the batch size "
            f"({batch_size}). {len(dataset) // batch_size} batches will be "
            f"constructed with a batch size of {batch_size}, and the last batch will "
            f"have a size of {len(dataset) % batch_size}. This might lead to "
            "inaccurate average timings."
        )

    # Create a dataloader
    dataloader = torch.utils.data.DataLoader(
        dataset,
        batch_size=batch_size,
        collate_fn=collate_fn,
        shuffle=False,
    )

    # Initialize RMSE accumulator:
    rmse_accumulator = RMSEAccumulator()
    mae_accumulator = MAEAccumulator()

    # If we're returning the predictions, we need to store them:
    if return_predictions:
        all_predictions = []

    # Set up timings:
    total_time = 0.0
    timings_per_atom = []

    # Warm up with a single batch 10 times (to get accurate timings later).
    # We use different batches to warm up torch potentially with different
    # tensor sizes, so dynamic shape compilation happens. We have to cycle
    # the dataloader in case there are few batches.
    cycled_dataloader = itertools.cycle(dataloader)
    for _ in range(10):
        batch = next(cycled_dataloader)
        systems = batch[0]
        systems = [system.to(dtype=dtype, device=device) for system in systems]
        evaluate_model(
            model,
            systems,
            options,
            is_training=False,
            check_consistency=check_consistency,
        )

    # Evaluate the model
    for batch in dataloader:
        systems, batch_targets = batch
        systems = [system.to(dtype=dtype, device=device) for system in systems]
        batch_targets = {
            key: value.to(dtype=dtype, device=device)
            for key, value in batch_targets.items()
        }

        start_time = time.time()

        batch_predictions = evaluate_model(
            model,
            systems,
            options,
            is_training=False,
            check_consistency=check_consistency,
        )

        if torch.cuda.is_available():
            torch.cuda.synchronize()
        end_time = time.time()

        batch_predictions_per_atom = average_by_num_atoms(
            batch_predictions, systems, per_structure_keys=[]
        )
        batch_targets_per_atom = average_by_num_atoms(
            batch_targets, systems, per_structure_keys=[]
        )
        rmse_accumulator.update(batch_predictions_per_atom, batch_targets_per_atom)
        mae_accumulator.update(batch_predictions_per_atom, batch_targets_per_atom)
        if return_predictions:
            all_predictions.append(batch_predictions)

        time_taken = end_time - start_time
        total_time += time_taken
        timings_per_atom.append(time_taken / sum(len(system) for system in systems))

    # Finalize the metrics
    rmse_values = rmse_accumulator.finalize(not_per_atom=["positions_gradients"])
    mae_values = mae_accumulator.finalize(not_per_atom=["positions_gradients"])
    metrics = {**rmse_values, **mae_values}

    # print the RMSEs with MetricLogger
    metric_logger = MetricLogger(
        log_obj=logger,
        dataset_info=model.capabilities(),
        initial_metrics=metrics,
    )
    metric_logger.log(metrics)

    # Log timings
    timings_per_atom = np.array(timings_per_atom)
    mean_per_atom = np.mean(timings_per_atom)
    std_per_atom = np.std(timings_per_atom)
    logging.info(
        f"evaluation time: {total_time:.2f} s "
        f"[{1000.0 * mean_per_atom:.4f} ± "
        f"{1000.0 * std_per_atom:.4f} ms per atom]"
    )

    if return_predictions:
        # concatenate the TensorMaps
        all_predictions_joined = _concatenate_tensormaps(all_predictions)
        return all_predictions_joined
    else:
        return None


def eval_model(
    model: Union[MetatensorAtomisticModel, torch.jit._script.RecursiveScriptModule],
    options: DictConfig,
    output: Union[Path, str] = "output.xyz",
    batch_size: int = 1,
    check_consistency: bool = False,
) -> None:
    """Evaluate an exported model on a given data set.

    If ``options`` contains a ``targets`` sub-section, RMSE values will be reported. If
    this sub-section is missing, only a xyz-file with containing the properties the
    model was trained against is written.

    :param model: Saved model to be evaluated.
    :param options: DictConfig to define a test dataset taken for the evaluation.
    :param output: Path to save the predicted values.
    :param check_consistency: Whether to run consistency checks during model evaluation.
    """
    logging.info("Setting up evaluation set.")

    if isinstance(output, str):
        output = Path(output)

    options_list = expand_dataset_config(options)
    for i, options in enumerate(options_list):
        if len(options_list) == 1:
            extra_log_message = ""
            file_index_suffix = ""
        else:
            extra_log_message = f" with index {i}"
            file_index_suffix = f"_{i}"
        logging.info(f"Evaluating dataset{extra_log_message}")

        eval_systems = read_systems(
            filename=options["systems"]["read_from"],
            reader=options["systems"]["reader"],
        )

        if hasattr(options, "targets"):
            # in this case, we only evaluate the targets specified in the options
            # and we calculate RMSEs
            eval_targets, eval_info_dict = read_targets(options["targets"])
        else:
            # in this case, we have no targets: we evaluate everything
            # (but we don't/can't calculate RMSEs)
            # TODO: allow the user to specify which outputs to evaluate
            eval_targets = {}
            eval_info_dict = {}
            do_strain_grad = all(
                not torch.all(system.cell == 0) for system in eval_systems
            )
            layout = _get_energy_layout(do_strain_grad)  # TODO: layout from the user
            for key in model.capabilities().outputs.keys():
                eval_info_dict[key] = TargetInfo(
                    quantity=model.capabilities().outputs[key].quantity,
                    unit=model.capabilities().outputs[key].unit,
                    # TODO: allow the user to specify whether per-atom or not
                    layout=layout,
                )

        eval_dataset = Dataset.from_dict({"system": eval_systems, **eval_targets})

        # Evaluate the model
        try:
            predictions = _eval_targets(
                model=model,
                dataset=eval_dataset,
                options=eval_info_dict,
                return_predictions=True,
                batch_size=batch_size,
                check_consistency=check_consistency,
            )
        except Exception as e:
            raise ArchitectureError(e)

        write_predictions(
            filename=f"{output.stem}{file_index_suffix}{output.suffix}",
            systems=eval_systems,
            capabilities=model.capabilities(),
            predictions=predictions,
        )


def _get_energy_layout(strain_gradient: bool) -> TensorMap:
    block = TensorBlock(
        # float64: otherwise metatensor can't serialize
        values=torch.empty(0, 1, dtype=torch.float64),
        samples=Labels(
            names=["system"],
            values=torch.empty((0, 1), dtype=torch.int32),
        ),
        components=[],
        properties=Labels.range("energy", 1),
    )
    position_gradient_block = TensorBlock(
        # float64: otherwise metatensor can't serialize
        values=torch.empty(0, 3, 1, dtype=torch.float64),
        samples=Labels(
            names=["sample", "atom"],
            values=torch.empty((0, 2), dtype=torch.int32),
        ),
        components=[
            Labels(
                names=["xyz"],
                values=torch.arange(3, dtype=torch.int32).reshape(-1, 1),
            ),
        ],
        properties=Labels.range("energy", 1),
    )
    block.add_gradient("positions", position_gradient_block)

    if strain_gradient:
        strain_gradient_block = TensorBlock(
            # float64: otherwise metatensor can't serialize
            values=torch.empty(0, 3, 3, 1, dtype=torch.float64),
            samples=Labels(
                names=["sample", "atom"],
                values=torch.empty((0, 2), dtype=torch.int32),
            ),
            components=[
                Labels(
                    names=["xyz_1"],
                    values=torch.arange(3, dtype=torch.int32).reshape(-1, 1),
                ),
                Labels(
                    names=["xyz_2"],
                    values=torch.arange(3, dtype=torch.int32).reshape(-1, 1),
                ),
            ],
            properties=Labels.range("energy", 1),
        )
        block.add_gradient("strain", strain_gradient_block)

    energy_layout = TensorMap(
        keys=Labels.single(),
        blocks=[block],
    )
    return energy_layout



"""Logging."""

import contextlib
import csv
import logging
import sys
from pathlib import Path
from typing import Dict, List, Optional, Tuple, Union

import numpy as np
import torch
from metatensor.torch.atomistic import ModelCapabilities

from .. import PACKAGE_ROOT, __version__
from .data import DatasetInfo
from .distributed.logging import is_main_process
from .external_naming import to_external_name
from .io import check_file_extension
from .units import ev_to_mev, get_gradient_units


try:
    from wandb.sdk.wandb_run import Run
except ImportError:
    Run = None


def _validate_length(keys: List[str], values: List[str], units: List[str]):
    if not (len(keys) == len(values) == len(units)):
        raise ValueError(
            f"keys, values and units must have the same length: "
            f"{len(keys)}, {len(values)}, {len(units)}"
        )


class CSVFileHandler(logging.FileHandler):
    """A custom FileHandler for logging data in CSV format."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._header_written = False

    def emit(self, record: logging.LogRecord):
        """Override the default behavior preventing any output to the default log."""
        pass

    def emit_data(self, keys: List[str], values: List[str], units: List[str]):
        """Write structured data to the CSV file.

        ``keys`` and ``values`` are written only the first time this methods is called.

        :param keys: Column header names
        :param values: Data values to write
        :param units: Units for each column
        """
        _validate_length(keys, values, units)

        with self._open() as file:
            writer = csv.writer(file)

            if not self._header_written:
                writer.writerow(keys)
                writer.writerow(units)
                self._header_written = True

            writer.writerow(values)


class WandbHandler(logging.Handler):
    """A custom logging handler that pushes structured logs to Weights & Biases.

    :param run: Weights & Biases run object.
    """

    def __init__(self, run: Run):
        super().__init__()
        self.run = run

    def emit(self, record: logging.LogRecord):
        """Override default behavior to ignore standard log records."""
        pass

    def emit_data(self, keys: List[str], values: List[str], units: List[str]):
        """Log structured data to Weights & Biases.

        :param keys: Column header names
        :param values: Data values to write
        :param units: Units for each column
        """
        _validate_length(keys, values, units)

        data = {}
        for key, value, unit in zip(keys, values, units):
            name = f"{key} [{unit}]" if unit else key
            data[name] = float(value)

        epoch = int(data.pop("Epoch"))
        self.run.log(data, step=epoch, commit=True)


class CustomLogger(logging.Logger):
    """Custom logger to log structured data."""

    def data(self, keys: List[str], values: List[str], units: List[str]):
        """Logs data entries to handlers that support an ``emit_data`` method.

        :param keys: Column header names
        :param values: Data values to write
        :param units: Units for each column
        """
        for handler in self.handlers:
            if hasattr(handler, "emit_data"):
                handler.emit_data(keys, values, units)


# Use `CustomLogger` as default class. The line below will NOT (!) change the `root`
# logger behaviour. It only applies for loggers that are created with a specific `name`
# like for example `logging.getLogger("metatrain")`. `logging.getLogger()` will still
# return the root logger (a `logging.Logger` instance).
logging.setLoggerClass(CustomLogger)


ROOT_LOGGER = logging.getLogger(name="metatrain")


class MetricLogger:
    def __init__(
        self,
        log_obj: Union[logging.Logger, CustomLogger],
        dataset_info: Union[ModelCapabilities, DatasetInfo],
        initial_metrics: Union[Dict[str, float], List[Dict[str, float]]],
        names: Union[str, List[str]] = "",
        scales: Optional[Dict[str, float]] = None,
    ):
        """
        Simple interface to log training metrics logging instance.

        Initialize the metric logger. The logger is initialized with the initial metrics
        and names relative to the metrics (e.g., "train", "validation").

        In this way, and by assuming that these metrics never increase, the logger can
        align the output to make it easier to read.

        :param log_obj: A logging instance
        :param model_outputs: outputs of the model. Used to infer physical quantities
            and units
        :param initial_metrics: initial training metrics
        :param names: names of the metrics (e.g., "train", "validation")
        """
        self.log_obj = log_obj

        # Length units will be used to infer units of forces/virials
        assert isinstance(dataset_info.length_unit, str)
        self.length_unit = dataset_info.length_unit

        # Save the model outputs. This will be useful to know
        # what physical quantities we are printing, along with their units
        if isinstance(dataset_info, DatasetInfo):
            self.model_outputs = dataset_info.targets
        elif isinstance(dataset_info, torch.ScriptObject):  # ModelCapabilities
            self.model_outputs = dataset_info.outputs
        else:
            raise ValueError(
                f"dataset_info must be of type `DatasetInfo` or `ModelCapabilities`, "
                f"not {type(dataset_info)}"
            )

        if isinstance(initial_metrics, dict):
            initial_metrics = [initial_metrics]
        if isinstance(names, str):
            names = [names]

        self.names = names

        if scales is None:
            scales = {target_name: 1.0 for target_name in initial_metrics[0].keys()}
        self.scales = scales

        # Since the quantities are supposed to decrease, we want to store the
        # number of digits at the start of the training, so that we can align
        # the output later:
        self.digits = {}
        for name, metrics_dict in zip(names, initial_metrics):
            for key, value in metrics_dict.items():
                value *= scales[key]
                target_name = key.split(" ", 1)[0]
                if key == "loss":
                    # losses will be printed in scientific notation
                    continue
                unit = self._get_units(target_name)
                value, unit = ev_to_mev(value, unit)
                self.digits[f"{name}_{key}"] = _get_digits(value)

    def log(
        self,
        metrics: Union[Dict[str, float], List[Dict[str, float]]],
        epoch: Optional[int] = None,
        rank: Optional[int] = None,
    ):
        """
        Log the metrics.

        The metrics are automatically aligned to make them easier to read, based on
        the order of magnitude of each metric given to the class at initialization.

        :param metrics: The current metrics to be logged.
        :param epoch: The current epoch (optional). If :py:class:`None`, the epoch
            will not be printed, and the logging string will start with the first
            metric in the ``metrics`` dictionary.
        :param rank: The rank of the process, if the training is distributed. In that
            case, the logger will only print the metrics for the process with rank 0.
        """
        if rank and rank != 0:
            return

        keys = []
        values = []
        units = []

        if epoch is not None:
            keys.append("Epoch")
            values.append(f"{epoch:4}")
            units.append("")

        if isinstance(metrics, dict):
            metrics = [metrics]

        is_loss = False
        for name, metrics_dict in zip(self.names, metrics):
            for key in _sort_metric_names(metrics_dict.keys()):
                value = metrics_dict[key] * self.scales[key]

                if key == "loss":
                    is_loss = True

                    # avoiding double spaces: only include non-empty strings (`if p`),
                    keys.append(" ".join(p for p in [name, key] if p))
                    values.append(f"{value:.3e}")
                    units.append("")

                else:  # special case: not a metric associated with a target
                    target_name, metric = key.split(" ", 1)
                    external_name = to_external_name(target_name, self.model_outputs)  # type: ignore # noqa: E501
                    keys.append(" ".join(p for p in [name, external_name, metric] if p))

                    unit = self._get_units(target_name)
                    value, unit = ev_to_mev(value, unit)

                    values.append(
                        f"{value:{self.digits[f'{name}_{key}'][0]}.{self.digits[f'{name}_{key}'][1]}f}"  # noqa: E501
                    )
                    units.append(unit)

        if is_loss and isinstance(self.log_obj, CustomLogger):
            self.log_obj.data(keys, values, units)

        # add space between value and unit only if the unit is not empty. Avoiding
        # double space when joining metric below
        formatted_metrics = [
            f"{key}: {value}{f' {unit}' if unit else ''}"
            for key, value, unit in zip(keys, values, units)
        ]

        logging.info(" | ".join(formatted_metrics))

    def _get_units(self, output: str) -> str:
        # Gets the units of an output
        if output.endswith("_gradients"):
            # handling <base_name>_<gradient_name>_gradients
            base_name = output[:-10]
            gradient_name = base_name.split("_")[-1]
            base_name = base_name.replace(f"_{gradient_name}", "")
            base_unit = self.model_outputs[base_name].unit
            unit = self._get_gradient_units(base_unit, gradient_name)
        else:
            unit = self.model_outputs[output].unit
        return unit

    def _get_gradient_units(self, base_unit: str, gradient_name: str) -> str:
        # Get the gradient units based on the unit of the base quantity
        # for example, if the base unit is "<unit>" and the gradient name is
        # "positions", the gradient unit will be "<unit>/<length_unit>".
        return get_gradient_units(base_unit, gradient_name, self.length_unit)


def _get_digits(value: float) -> Tuple[int, int]:
    """
    Finds the number of digits to print before and after the decimal point,
    based on the order of magnitude of the value.

    5 "significant" digits are guaranteed to be printed.

    :param value: The value for which the number of digits is calculated.
    """

    # Get order of magnitude of the value:
    order = int(np.floor(np.log10(value)))

    # Get the number of digits before the decimal point:
    if order < 0:
        digits_before = 1
    else:
        digits_before = order + 1

    # Get the number of digits after the decimal point:
    if order < 0:
        digits_after = 4 - order
    else:
        digits_after = max(1, 4 - order)

    total_characters = digits_before + digits_after + 1  # +1 for the point

    return total_characters, digits_after


@contextlib.contextmanager
def setup_logging(
    log_obj: logging.Logger,
    log_file: Optional[Union[str, Path]] = None,
    level: int = logging.WARNING,
):
    """Create a logging environment for a given ``log_obj``.

    Extracted and adjusted from
    github.com/MDAnalysis/mdacli/blob/main/src/mdacli/logger.py

    :param log_obj: A logging instance
    :param log_file: Name of the log file
    :param level: Set the root logger level to the specified level. If for example set
        to :py:obj:`logging.DEBUG` detailed debug logs including filename and function
        name are displayed. For :py:obj:`logging.INFO` only the message logged from
        `errors`, `warnings` and `infos` will be displayed.
    """
    try:
        format = "[{asctime}][{levelname}]"
        if level == logging.DEBUG:
            format += ":{filename}:{funcName}:{lineno}"
        format += " - {message}"

        formatter = logging.Formatter(format, datefmt="%Y-%m-%d %H:%M:%S", style="{")
        handlers: List[Union[logging.StreamHandler, logging.FileHandler]] = []

        stream_handler = logging.StreamHandler(sys.stdout)
        stream_handler.setFormatter(formatter)
        handlers.append(stream_handler)

        if log_file and is_main_process():
            log_file = check_file_extension(filename=log_file, extension=".log")
            file_handler = logging.FileHandler(filename=str(log_file), encoding="utf-8")
            file_handler.setFormatter(formatter)
            handlers.append(file_handler)

            csv_file = Path(log_file).with_suffix(".csv")
            csv_handler = CSVFileHandler(str(csv_file))
            handlers.append(csv_handler)

        # hide logging up to ERROR from secondary processes in distributed environments:
        if not is_main_process():
            level = logging.ERROR

        # set the level for root logger
        logging.basicConfig(format=format, handlers=handlers, level=level, style="{")
        logging.captureWarnings(True)

        if log_file:
            abs_path = str(Path(log_file).absolute().resolve())
            log_obj.info(f"This log is also available at {abs_path!r}.")
        else:
            log_obj.info("Logging to file is disabled.")

        log_obj.info(f"Package version: {__version__}")
        log_obj.info(f"Package directory: {PACKAGE_ROOT}")
        log_obj.info(f"Working directory: {Path('.').absolute()}")
        log_obj.info(f"Executed command: {get_cli_input()}")

        # keep in the end to avoid double logging
        for handler in handlers:
            log_obj.addHandler(handler)

        yield

    finally:
        for handler in handlers:
            handler.flush()
            handler.close()

            if isinstance(handler, WandbHandler):
                handler.run.finish()

            log_obj.removeHandler(handler)


def get_cli_input(argv: Optional[List[str]] = None) -> str:
    """Proper formatted string of the command line input.

    :param argv: List of strings to parse. If :py:obj:`None` taken from
        :py:obj:`sys.argv`.
    """
    if argv is None:
        argv = sys.argv

    program_name = Path(argv[0]).name
    # Add additional quotes for connected arguments.
    arguments = [f'"{arg}"' if " " in arg else arg for arg in argv[1:]]
    return f"{program_name} {' '.join(arguments)}"


def _sort_metric_names(name_list):
    name_list = list(name_list)
    sorted_name_list = []
    if "loss" in name_list:
        # loss goes first
        loss_index = name_list.index("loss")
        sorted_name_list.append(name_list.pop(loss_index))
    # then alphabetical order, except for the MAEs, which should come
    # after the corresponding RMSEs
    sorted_remaining_name_list = sorted(
        name_list,
        key=lambda x: x.replace("RMSE", "AAA").replace("MAE", "ZZZ"),
    )
    # add the rest
    sorted_name_list.extend(sorted_remaining_name_list)
    return sorted_name_list



import torch


class FeedForwardBlock(torch.nn.Module):
    """A single transformer feed forward block."""

    def __init__(
        self,
        hidden_size: int,
        intermediate_size: int,
        dropout_rate: float,
    ):
        super().__init__()

        self.mlp = torch.nn.Linear(
            in_features=hidden_size, out_features=intermediate_size, bias=False
        )
        self.output = torch.nn.Linear(
            in_features=intermediate_size, out_features=hidden_size, bias=False
        )

        self.layernorm = torch.nn.LayerNorm(normalized_shape=hidden_size)
        self.dropout = torch.nn.Dropout(dropout_rate)

    def forward(
        self,
        inputs: torch.Tensor,  # hidden_size
    ) -> torch.Tensor:  # hidden_size
        # Pre-layer normalization
        normed_inputs = self.layernorm(inputs)

        # Feed-forward
        hidden = self.mlp(normed_inputs)
        hidden = torch.nn.functional.gelu(hidden)

        # Project back to input size
        outputs = self.output(hidden)

        # Apply dropout
        outputs = self.dropout(outputs)

        # Residual connection
        outputs = (outputs + inputs) * 0.5**0.5

        return outputs



import torch
from metatensor.torch.atomistic import ModelMetadata
from omegaconf import OmegaConf

from metatrain.gap import GAP, Trainer
from metatrain.utils.data import Dataset, DatasetInfo
from metatrain.utils.data.readers import read_systems, read_targets
from metatrain.utils.data.target_info import get_energy_target_info

from . import DATASET_PATH, DEFAULT_HYPERS


def test_export():
    """Tests that export works with injected metadata"""

    systems = read_systems(DATASET_PATH)

    conf = {
        "energy": {
            "quantity": "energy",
            "read_from": DATASET_PATH,
            "reader": "ase",
            "key": "U0",
            "unit": "kcal/mol",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": False,
            "stress": False,
            "virial": False,
        }
    }
    targets, _ = read_targets(OmegaConf.create(conf))
    dataset = Dataset.from_dict({"system": systems, "energy": targets["energy"]})

    target_info_dict = {}
    target_info_dict["energy"] = get_energy_target_info({"unit": "eV"})

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[1, 6, 7, 8], targets=target_info_dict
    )

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )
    model = GAP(DEFAULT_HYPERS["model"], dataset_info)

    # we have to train gap before we can export...
    trainer = Trainer(DEFAULT_HYPERS["training"])
    trainer.train(
        model=model,
        dtype=torch.float64,
        devices=[torch.device("cpu")],
        train_datasets=[dataset],
        val_datasets=[dataset],
        checkpoint_dir=".",
    )

    exported = model.export(metadata=ModelMetadata(name="test"))

    # test correct metadata
    assert "This is the test model" in str(exported.metadata())



import numpy as np
import torch
from sklearn.linear_model import Ridge

from .molecule import Molecule


def get_all_species(structures):
    all_species = []
    for structure in structures:
        all_species.append(np.array(structure.get_atomic_numbers()))
    all_species = np.concatenate(all_species, axis=0)
    all_species = np.sort(np.unique(all_species))
    return all_species


def get_forces(structures, FORCES_KEY):
    forces = []
    for structure in structures:
        forces.append(
            torch.tensor(structure.arrays[FORCES_KEY], dtype=torch.get_default_dtype())
        )
    return forces


def update_pyg_graphs(pyg_graphs, key, values):
    for index in range(len(pyg_graphs)):
        pyg_graphs[index].update({key: values[index]})


def get_pyg_graphs(
    structures,
    all_species,
    R_CUT,
    USE_ADDITIONAL_SCALAR_ATTRIBUTES,
    USE_LONG_RANGE,
    K_CUT,
    MULTI_TARGET,
    TARGET_INDEX_KEY,
):
    molecules = [
        Molecule(
            structure,
            R_CUT,
            USE_ADDITIONAL_SCALAR_ATTRIBUTES,
            USE_LONG_RANGE,
            K_CUT,
            MULTI_TARGET,
            TARGET_INDEX_KEY,
        )
        for structure in structures
    ]

    max_nums = [molecule.get_max_num() for molecule in molecules]
    max_num = np.max(max_nums)

    if USE_LONG_RANGE:
        k_nums = [molecule.get_num_k() for molecule in molecules]
        max_k_num = np.max(k_nums)
    else:
        max_k_num = None

    pyg_graphs = [
        molecule.get_graph(max_num, all_species, max_k_num) for molecule in molecules
    ]
    return pyg_graphs


def get_compositional_features(structures, all_species):
    result = np.zeros([len(structures), len(all_species)])
    for i, structure in enumerate(structures):
        species_now = structure.get_atomic_numbers()
        for j, specie in enumerate(all_species):
            num = np.sum(species_now == specie)
            result[i, j] = num
    return result


def get_targets(structures, GENERAL_TARGET_SETTINGS):
    """Get general targets from structures"""

    targets = []
    if GENERAL_TARGET_SETTINGS.TARGET_TYPE not in ["structural", "atomic"]:
        raise ValueError("unknown target type")

    for structure in structures:
        if GENERAL_TARGET_SETTINGS.TARGET_TYPE == "structural":
            target_now = structure.info[GENERAL_TARGET_SETTINGS.TARGET_KEY]
            if not isinstance(target_now, np.ndarray):
                raise ValueError("target must be numpy array")
            if len(target_now.shape) != 1:
                raise ValueError("structural target must be 1D array")

            target_now = target_now[np.newaxis]

        elif GENERAL_TARGET_SETTINGS.TARGET_TYPE == "atomic":
            target_now = structure.arrays[GENERAL_TARGET_SETTINGS.TARGET_KEY]
            if not isinstance(target_now, np.ndarray):
                raise ValueError("target must be numpy array")
            if len(target_now.shape) != 2:
                raise ValueError("atomic target must be 2D array")

        targets.append(torch.tensor(target_now, dtype=torch.get_default_dtype()))
    return targets


def get_self_contributions(energy_key, train_structures, all_species):
    train_energies = np.array(
        [structure.info[energy_key] for structure in train_structures]
    )
    train_c_feat = get_compositional_features(train_structures, all_species)
    rgr = Ridge(alpha=1e-10, fit_intercept=False)
    rgr.fit(train_c_feat, train_energies)
    return rgr.coef_


def get_corrected_energies(energy_key, structures, all_species, self_contributions):
    energies = np.array([structure.info[energy_key] for structure in structures])

    compositional_features = get_compositional_features(structures, all_species)
    self_contributions_energies = []
    for i in range(len(structures)):
        self_contributions_energies.append(
            np.dot(compositional_features[i], self_contributions)
        )
    self_contributions_energies = np.array(self_contributions_energies)
    return energies - self_contributions_energies



from typing import List

import torch
from metatensor.torch.atomistic import NeighborListOptions, System


def concatenate_structures(
    systems: List[System], neighbor_list_options: NeighborListOptions
):
    positions = []
    centers = []
    neighbors = []
    species = []
    cell_shifts = []
    cells = []
    node_counter = 0

    for system in systems:
        positions.append(system.positions)
        species.append(system.types)

        assert len(system.known_neighbor_lists()) >= 1, "no neighbor list found"
        neighbor_list = system.get_neighbor_list(neighbor_list_options)
        nl_values = neighbor_list.samples.values

        centers.append(nl_values[:, 0] + node_counter)
        neighbors.append(nl_values[:, 1] + node_counter)
        cell_shifts.append(nl_values[:, 2:])

        cells.append(system.cell)

        node_counter += len(system.positions)

    positions = torch.cat(positions)
    centers = torch.cat(centers)
    neighbors = torch.cat(neighbors)
    species = torch.cat(species)
    cells = torch.stack(cells)
    cell_shifts = torch.cat(cell_shifts)

    return (
        positions,
        centers,
        neighbors,
        species,
        cells,
        cell_shifts,
    )



import torch


def get_radial_mask(r, r_cut: float, r_transition: float):
    # All radii are already guaranteed to be smaller than r_cut
    return torch.where(
        r < r_transition,
        torch.ones_like(r),
        0.5 * (torch.cos(torch.pi * (r - r_transition) / (r_cut - r_transition)) + 1.0),
    )



import pytest


pytest.importorskip("torchpme")

import copy

import torch
from metatensor.torch.atomistic import ModelOutput, System
from omegaconf import OmegaConf

from metatrain.experimental.nativepet import NativePET, Trainer
from metatrain.utils.data import Dataset, DatasetInfo
from metatrain.utils.data.readers import (
    read_systems,
    read_targets,
)
from metatrain.utils.data.target_info import (
    get_energy_target_info,
)
from metatrain.utils.evaluate_model import evaluate_model
from metatrain.utils.neighbor_lists import get_system_with_neighbor_lists

from . import DATASET_WITH_FORCES_PATH, DEFAULT_HYPERS, MODEL_HYPERS


@pytest.mark.parametrize("use_ewald", [True, False])
def test_long_range_features(use_ewald):
    """Tests that long-range features can be computed."""
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "energy": get_energy_target_info({"quantity": "energy", "unit": "eV"})
        },
    )
    hypers = copy.deepcopy(MODEL_HYPERS)
    hypers["long_range"]["enable"] = True
    hypers["long_range"]["use_ewald"] = use_ewald
    model = NativePET(hypers, dataset_info)

    system = System(
        types=torch.tensor([6, 6, 8, 8]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]]
        ),
        cell=torch.eye(3) * 10,
        pbc=torch.tensor([True, True, True]),
    )
    system = get_system_with_neighbor_lists(system, model.requested_neighbor_lists())
    outputs = {"energy": ModelOutput(per_atom=False)}
    model([system, system], outputs)


@pytest.mark.parametrize("use_ewald", [True, False])
def test_long_range_training(use_ewald):
    """Tests that long-range features can be computed."""
    pytest.importorskip("torch", minversion="1.20")
    systems = read_systems(DATASET_WITH_FORCES_PATH)

    conf = {
        "energy": {
            "quantity": "energy",
            "read_from": DATASET_WITH_FORCES_PATH,
            "reader": "ase",
            "key": "energy",
            "unit": "eV",
            "type": "scalar",
            "per_atom": False,
            "num_subtargets": 1,
            "forces": {"read_from": DATASET_WITH_FORCES_PATH, "key": "force"},
            "stress": False,
            "virial": False,
        }
    }

    targets, target_info_dict = read_targets(OmegaConf.create(conf))
    targets = {"energy": targets["energy"]}
    dataset = Dataset.from_dict({"system": systems, "energy": targets["energy"]})
    hypers = DEFAULT_HYPERS.copy()
    hypers["training"]["num_epochs"] = 2
    hypers["training"]["scheduler_patience"] = 1
    hypers["training"]["fixed_composition_weights"] = {}

    dataset_info = DatasetInfo(
        length_unit="Angstrom", atomic_types=[6], targets=target_info_dict
    )

    model_hypers = copy.deepcopy(MODEL_HYPERS)
    model_hypers["long_range"]["enable"] = True
    model_hypers["long_range"]["use_ewald"] = use_ewald
    model = NativePET(model_hypers, dataset_info)

    trainer = Trainer(hypers["training"])
    trainer.train(
        model=model,
        dtype=torch.float32,
        devices=[torch.device("cpu")],
        train_datasets=[dataset],
        val_datasets=[dataset],
        checkpoint_dir=".",
    )

    # Predict on the first five systems
    systems = [system.to(torch.float32) for system in systems]
    for system in systems:
        get_system_with_neighbor_lists(system, model.requested_neighbor_lists())

    evaluate_model(model, systems[:5], targets=target_info_dict, is_training=False)



from typing import Dict, List

import ase
import ase.io
import metatensor.torch
import torch
from metatensor.torch import Labels, TensorMap
from metatensor.torch.atomistic import ModelCapabilities, System

from ...external_naming import to_external_name


def write_xyz(
    filename: str,
    systems: List[System],
    capabilities: ModelCapabilities,
    predictions: Dict[str, TensorMap],
) -> None:
    """An ase-based xyz file writer. Writes the systems and predictions to an xyz file.

    According to ASE practice, arrays which have a dimension corresponding
    to each atom are saved inside atoms.arrays, while any other arrays are
    saved inside atoms.info.

    :param filename: name of the file to save to.
    :param systems: structures to be written to the file.
    :param: capabilities: capabilities of the model.
    :param predictions: prediction values to be written to the file.
    """

    # we first split the predictions by structure
    predictions_by_structure: List[Dict[str, TensorMap]] = [{} for _ in systems]
    split_labels = [
        Labels(names=["system"], values=torch.tensor([[i_system]]))
        for i_system in range(len(systems))
    ]
    for target_name, target_tensor_map in predictions.items():
        # split this target by structure
        target_tensor_map = target_tensor_map.to("cpu")
        split_target = metatensor.torch.split(
            target_tensor_map, "samples", split_labels
        )
        for i_system, system_target in enumerate(split_target):
            # add the split target to the dict corresponding to the structure
            predictions_by_structure[i_system][target_name] = system_target

    frames = []
    for system, system_predictions in zip(systems, predictions_by_structure):
        info = {}
        arrays = {}
        for target_name, target_map in system_predictions.items():
            if len(target_map.keys) != 1:
                raise ValueError(
                    "Only single-block `TensorMap`s can be "
                    "written to xyz files for the moment."
                )
            block = target_map.block()
            if "atom" in block.samples.names:
                # save inside arrays
                values = block.values.detach().cpu().numpy()
                arrays[target_name] = values.reshape(values.shape[0], -1)
                # reshaping here is necessary because `arrays` only accepts 2D arrays
            else:
                # save inside info
                if block.values.numel() == 1:
                    info[target_name] = block.values.item()
                else:
                    info[target_name] = block.values.detach().cpu().numpy().squeeze(0)
                    # squeeze the sample dimension, which corresponds to the system

            for gradient_name, gradient_block in block.gradients():
                # here, we assume that gradients are always an array, and never a scalar
                internal_name = f"{target_name}_{gradient_name}_gradients"
                external_name = to_external_name(internal_name, capabilities.outputs)

                if "forces" in external_name:
                    arrays[external_name] = (
                        # squeeze the property dimension
                        -gradient_block.values.detach().cpu().squeeze(-1).numpy()
                    )
                elif "virial" in external_name:
                    # in this case, we write both the virial and the stress
                    external_name_virial = external_name
                    external_name_stress = external_name.replace("virial", "stress")
                    strain_derivatives = (
                        # squeeze the property dimension
                        gradient_block.values.detach().cpu().squeeze(-1).numpy()
                    )
                    if not torch.any(system.cell != 0):
                        raise ValueError(
                            "stresses cannot be written for non-periodic systems."
                        )
                    cell_volume = torch.det(system.cell).item()
                    if cell_volume == 0:
                        raise ValueError(
                            "stresses cannot be written for systems with zero volume."
                        )
                    info[external_name_virial] = -strain_derivatives
                    info[external_name_stress] = strain_derivatives / cell_volume
                else:
                    info[external_name] = (
                        # squeeze the property dimension
                        gradient_block.values.detach().cpu().squeeze(-1).numpy()
                    )

        atoms = ase.Atoms(
            symbols=system.types, positions=system.positions.detach(), info=info
        )

        # assign cell and pbcs
        if torch.any(system.cell != 0):
            atoms.pbc = True
            atoms.cell = system.cell.detach().cpu().numpy()

        # assign arrays
        for array_name, array in arrays.items():
            atoms.arrays[array_name] = array

        frames.append(atoms)

    ase.io.write(filename, frames)



import difflib

import jsonschema
from jsonschema.exceptions import ValidationError


def validate(instance, schema, cls=None, *args, **kwargs) -> None:
    """Validate an instance under the given schema.

    Function similar to :py:class:`jsonschema.validate` but displaying only the human
    readable error message without showing the reference schema and path if the instance
    is invalid. In addition, if the error is caused by unallowed
    ``additionalProperties`` the closest matching properties will be suggested.

    :param instance: Instance to validate
    :param schema: Schema to validate with
    :raises jsonschema.exceptions.ValidationError: If the instance is invalid
    :raises jsonschema.exceptions.SchemaError: If the schema itself is invalid
    """
    try:
        jsonschema.validate(instance, schema, cls=cls, *args, **kwargs)  # noqa: B026
    except ValidationError as error:
        if error.validator == "additionalProperties":
            # Change error message to be clearer for users
            error.message = error.message.replace(
                "Additional properties are not allowed", "Unrecognized options"
            )

            known_properties = error.schema["properties"].keys()
            unknown_properties = error.instance.keys() - known_properties

            closest_matches = []
            for name in unknown_properties:
                closest_match = difflib.get_close_matches(
                    word=name, possibilities=known_properties
                )

                if closest_match:
                    closest_matches.append(f"'{closest_match[0]}'")

            if closest_matches:
                error.message += f". Do you mean {', '.join(closest_matches)}?"

        raise ValidationError(message=error.message)



from typing import Dict

import metatensor.torch
import torch
from metatensor.torch import Labels, TensorBlock, TensorMap


def convert_model_state_dict_from_legacy_pet(
    checkpoint: Dict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    """Converts the model state dict from the metatain.pet format
    to the metatrain.experimental.nativepet format.
    """

    new_model_state_dict: Dict[str, torch.Tensor] = {}
    model_state_dict = checkpoint["model_state_dict"]
    for key, value in model_state_dict.items():
        new_key = key.replace("module.model.model.pet_model.", "")
        if "embedding" in new_key and "gnn_layers" not in new_key:
            new_model_state_dict[new_key] = value
        if "gnn_layers" in new_key and "trans_layer" not in new_key:
            if "r_embedding" in new_key:
                new_key = new_key.replace("r_embedding", "edge_embedder")
            if "central_embedder" in new_key:
                new_key = new_key.replace("central_embedder", "node_embedder")
            new_model_state_dict[new_key] = value

        if "heads" in new_key:
            if "linear" in new_key:
                layer_num = int(new_key.split(".")[1])
                if "bond" in new_key:
                    last_layer_key = f"edge_last_layers.energy.{layer_num}.energy___0"
                else:
                    last_layer_key = f"node_last_layers.energy.{layer_num}.energy___0"
                if "weight" in new_key:
                    last_layer_key += ".weight"
                if "bias" in new_key:
                    last_layer_key += ".bias"
                new_model_state_dict[last_layer_key] = value
            else:
                if "bond" in new_key:
                    new_key = new_key.replace("bond_heads.", "edge_heads.energy.")
                else:
                    new_key = new_key.replace("heads.", "node_heads.energy.")
                new_key = new_key.replace("nn.", "")
                new_model_state_dict[new_key] = value

    new_model_state_dict["scaler.scales"] = torch.tensor([1.0], dtype=torch.float32)
    new_model_state_dict["additive_models.0.dummy_buffer"] = torch.tensor(
        [1.0], dtype=torch.float64
    )

    atomic_types = torch.tensor(checkpoint["dataset_info"].atomic_types)

    samples = Labels(names=["center_type"], values=atomic_types.unsqueeze(1))
    weights = TensorMap(
        keys=Labels.range("_", 1),
        blocks=[
            TensorBlock(
                samples=samples,
                components=[],
                properties=Labels.range("energy", 1),
                values=torch.tensor(
                    checkpoint["self_contributions"], dtype=torch.float64
                ).unsqueeze(1),
            )
        ],
    )

    new_model_state_dict["additive_models.0.energy_composition_buffer"] = (
        metatensor.torch.save_buffer(weights)
    )

    species_to_species_index = torch.full(
        (max(atomic_types) + 1,),
        -1,
        dtype=torch.int64,
    )
    for i, species in enumerate(atomic_types):
        species_to_species_index[species] = i

    new_model_state_dict["species_to_species_index"] = species_to_species_index
    new_model_state_dict["additive_models.0.type_to_index"] = species_to_species_index
    return new_model_state_dict


def convert_model_data_from_legacy_pet(
    checkpoint: Dict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    """Converts the model data from the metatain.pet format
    to the metatrain.experimental.nativepet format.
    """

    new_model_data: Dict[str, torch.Tensor] = {}
    new_model_data["dataset_info"] = checkpoint["dataset_info"]
    hypers = checkpoint["hypers"]["ARCHITECTURAL_HYPERS"]
    new_model_data["model_hypers"] = {
        "cutoff": hypers["R_CUT"],
        "cutoff_width": hypers["CUTOFF_DELTA"],
        "d_pet": hypers["TRANSFORMER_D_MODEL"],
        "d_head": hypers["HEAD_N_NEURONS"],
        "d_feedforward": hypers["TRANSFORMER_DIM_FEEDFORWARD"],
        "num_heads": hypers["TRANSFORMER_N_HEAD"],
        "num_attention_layers": hypers["N_TRANS_LAYERS"],
        "num_gnn_layers": hypers["N_GNN_LAYERS"],
        "residual_factor": hypers["RESIDUAL_FACTOR"],
        "zbl": hypers["USE_ZBL"],
        "long_range": {
            "enable": False,
            "use_ewald": False,
            "smearing": 1.4,
            "kspace_resolution": 1.33,
            "interpolation_nodes": 5,
        },
    }

    return new_model_data


def convert_train_hypers_from_legacy_pet(
    checkpoint: Dict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    """Converts the training hypers from the metatain.pet format
    to the metatrain.experimental.nativepet format.
    """
    hypers = checkpoint["hypers"]["FITTING_SCHEME"]
    new_train_hypers: Dict[str, torch.Tensor] = {
        "distributed": False,
        "distributed_port": 39591,
        "batch_size": 16,
        "num_epochs": hypers["EPOCH_NUM"],
        "num_epochs_warmup": hypers["EPOCHS_WARMUP"],
        "learning_rate": hypers["INITIAL_LR"],
        "weight_decay": hypers["WEIGHT_DECAY"] if hypers["USE_WEIGHT_DECAY"] else None,
        "scheduler_patience": hypers["SCHEDULER_STEP_SIZE"],
        "log_interval": 1,
        "checkpoint_interval": hypers["CHECKPOINT_INTERVAL"],
        "scale_targets": False,
        "fixed_composition_weights": {},
        "per_structure_targets": [],
        "log_mae": True,
        "log_separate_blocks": False,
        "best_model_metric": "rmse_prod",
        "loss": {
            "type": "mse",
            "weights": {"energy": hypers["ENERGY_WEIGHT"]},
            "reduction": "mean",
            "sliding_factor": hypers["SLIDING_FACTOR"],
        },
    }

    return new_train_hypers


def convert_checkpoint_from_legacy_pet(
    checkpoint: Dict[str, torch.Tensor],
) -> Dict[str, torch.Tensor]:
    """Converts the state dict from the metatain.pet format
    to the metatrain.experimental.nativepet format.
    """

    new_checkpoint: Dict[str, torch.Tensor] = {}
    new_checkpoint["architecture_name"] = "experimental.nativepet"
    new_checkpoint["epoch"] = 0
    new_checkpoint["optimizer_state_dict"] = None
    new_checkpoint["scheduler_state_dict"] = None
    new_checkpoint["best_metric"] = float("inf")
    new_checkpoint["best_model_state_dict"] = None
    new_checkpoint["best_optimizer_state_dict"] = None
    new_checkpoint["model_state_dict"] = convert_model_state_dict_from_legacy_pet(
        checkpoint
    )
    new_checkpoint["model_data"] = convert_model_data_from_legacy_pet(checkpoint)
    new_checkpoint["train_hypers"] = convert_train_hypers_from_legacy_pet(checkpoint)

    return new_checkpoint



from pathlib import Path

from metatrain.utils.architectures import get_default_hypers


DEFAULT_HYPERS = get_default_hypers("gap")
DATASET_PATH = str(Path(__file__).parents[4] / "tests/resources/qm9_reduced_100.xyz")

DATASET_ETHANOL_PATH = str(
    Path(__file__).parents[4] / "tests/resources/ethanol_reduced_100.xyz"
)



from typing import Dict, List

import torch
from metatensor.torch import TensorMap
from metatensor.torch.atomistic import System


@torch.jit.script
def systems_and_targets_to_device(  # pragma: no cover
    systems: List[System],
    targets: Dict[str, TensorMap],
    device: torch.device,
):
    """
    Transfers the systems and targets to the specified device.

    :param systems: List of systems.
    :param targets: Dictionary of targets.
    :param device: Device to transfer to.
    """

    systems = [system.to(device=device) for system in systems]
    targets = {key: value.to(device=device) for key, value in targets.items()}
    return systems, targets


@torch.jit.script
def systems_and_targets_to_dtype(  # pragma: no cover
    systems: List[System],
    targets: Dict[str, TensorMap],
    dtype: torch.dtype,
):
    """
    Changes the systems and targets to the specified floating point data type.

    :param systems: List of systems.
    :param targets: Dictionary of targets.
    :param dtype: Desired floating point data type.
    """

    systems = [system.to(dtype=dtype) for system in systems]
    targets = {key: value.to(dtype=dtype) for key, value in targets.items()}
    return systems, targets



import pytest
import torch
from metatensor.torch.atomistic import ModelEvaluationOptions, ModelMetadata, System

from metatrain.soap_bpnn import SoapBpnn
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import get_energy_target_info
from metatrain.utils.neighbor_lists import (
    get_requested_neighbor_lists,
    get_system_with_neighbor_lists,
)

from . import MODEL_HYPERS


@pytest.mark.parametrize("device", ["cpu", "cuda"])
@pytest.mark.parametrize("dtype", [torch.float32, torch.float64])
def test_to(device, dtype):
    """Tests that the `.to()` method of the exported model works."""
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA is not available")

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = SoapBpnn(MODEL_HYPERS, dataset_info).to(dtype=dtype)
    exported = model.export(metadata=ModelMetadata(name="test"))

    # test correct metadata
    assert "This is the test model" in str(exported.metadata())

    exported.to(device=device)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    requested_neighbor_lists = get_requested_neighbor_lists(exported)
    system = get_system_with_neighbor_lists(system, requested_neighbor_lists)
    system = system.to(device=device, dtype=dtype)

    evaluation_options = ModelEvaluationOptions(
        length_unit=dataset_info.length_unit,
        outputs=model.outputs,
    )

    exported([system], evaluation_options, check_consistency=True)



import warnings
from typing import List, Optional

import torch


def compute_gradient(
    target: torch.Tensor, inputs: List[torch.Tensor], is_training: bool
) -> List[torch.Tensor]:
    """
    Calculates the gradient of a target tensor with respect to a list of input tensors.

    ``target`` must be a single torch.Tensor object. If target contains multiple values,
    the gradient will be calculated with respect to the sum of all values.
    """

    grad_outputs: Optional[List[Optional[torch.Tensor]]] = [torch.ones_like(target)]
    try:
        gradient = torch.autograd.grad(
            outputs=[target],
            inputs=inputs,
            grad_outputs=grad_outputs,
            retain_graph=is_training,
            create_graph=is_training,
        )
    except RuntimeError as e:
        # Torch raises an error if the target tensor does not require grad,
        # but this could just mean that the target is a constant tensor, like in
        # the case of composition models. In this case, we can safely ignore the error
        # and we raise a warning instead. The warning can be caught and silenced in the
        # appropriate places.
        if (
            "element 0 of tensors does not require grad and does not have a grad_fn"
            in str(e)
        ):
            warnings.warn(f"GRADIENT WARNING: {e}", RuntimeWarning, stacklevel=2)
            gradient = [torch.zeros_like(i) for i in inputs]
        else:
            # Re-raise the error if it's not the one above
            raise
    if gradient is None:
        raise ValueError(
            "Unexpected None value for computed gradient. "
            "One or more operations inside the model might "
            "not have a gradient implementation."
        )
    else:
        return gradient



import torch
import torch.nn.functional as F
from torch import nn

from .utilities import DummyModule


class AttentionBlock(nn.Module):
    def __init__(self, total_dim, num_heads, epsilon=1e-15):
        super(AttentionBlock, self).__init__()

        self.input_linear = nn.Linear(total_dim, 3 * total_dim)
        self.output_linear = nn.Linear(total_dim, total_dim)

        nn.init.xavier_uniform_(self.input_linear.weight)
        nn.init.constant_(self.input_linear.bias, 0.0)
        nn.init.constant_(self.output_linear.bias, 0.0)

        self.num_heads = num_heads
        self.epsilon = epsilon

        if total_dim % num_heads != 0:
            raise ValueError("total dimension is not divisible by the number of heads")
        self.head_dim = total_dim // num_heads

    def forward(self, x, cutoff_factors: torch.Tensor, use_manual_attention: bool):
        initial_shape = x.shape
        x = self.input_linear(x)
        x = x.reshape(
            initial_shape[0], initial_shape[1], 3, self.num_heads, self.head_dim
        )
        x = x.permute(2, 0, 3, 1, 4)

        queries, keys, values = x[0], x[1], x[2]
        attn_weights = torch.clamp(cutoff_factors[:, None, :, :], self.epsilon)
        attn_weights = torch.log(attn_weights)
        if use_manual_attention:
            x = manual_attention(queries, keys, values, attn_weights)
        else:
            x = torch.nn.functional.scaled_dot_product_attention(
                queries,
                keys,
                values,
                attn_mask=attn_weights,
            )
        x = x.transpose(1, 2).reshape(initial_shape)
        x = self.output_linear(x)
        return x


class TransformerLayer(torch.nn.Module):
    def __init__(
        self,
        d_model,
        n_heads,
        dim_feedforward=512,
        dropout=0.0,
        activation=F.silu,
        transformer_type="PostLN",
    ):
        super(TransformerLayer, self).__init__()
        self.attention = AttentionBlock(d_model, n_heads)

        if transformer_type not in ["PostLN", "PreLN"]:
            raise ValueError("unknown transformer type")
        self.transformer_type = transformer_type
        self.d_model = d_model
        self.norm_attention = nn.LayerNorm(d_model)
        self.norm_mlp = nn.LayerNorm(d_model)
        self.dropout = nn.Dropout(dropout)

        self.activation = activation

        self.mlp = nn.Sequential(
            nn.Linear(d_model, dim_feedforward),
            self.activation,
            nn.Dropout(dropout),
            nn.Linear(dim_feedforward, d_model),
            nn.Dropout(dropout),
        )

    def forward(
        self,
        tokens: torch.Tensor,
        cutoff_factors: torch.Tensor,
        use_manual_attention: bool,
    ) -> torch.Tensor:
        if self.transformer_type == "PostLN":
            tokens = self.norm_attention(
                tokens
                + self.dropout(
                    self.attention(tokens, cutoff_factors, use_manual_attention)
                )
            )
            tokens = self.norm_mlp(tokens + self.mlp(tokens))
        if self.transformer_type == "PreLN":
            tokens = tokens + self.dropout(
                self.attention(
                    self.norm_attention(tokens), cutoff_factors, use_manual_attention
                )
            )
            tokens = tokens + self.mlp(self.norm_mlp(tokens))
        return tokens


class Transformer(torch.nn.Module):
    def __init__(
        self,
        d_model,
        num_layers,
        n_heads,
        dim_feedforward=512,
        dropout=0.0,
        activation=F.silu,
        transformer_type="PostLN",
    ):
        super(Transformer, self).__init__()
        self.transformer_type = transformer_type

        self.final_norm = DummyModule()  # for torchscript
        if transformer_type == "PreLN":
            self.final_norm = nn.LayerNorm(d_model)
        self.layers = nn.ModuleList(
            [
                TransformerLayer(
                    d_model=d_model,
                    n_heads=n_heads,
                    dim_feedforward=dim_feedforward,
                    dropout=dropout,
                    activation=activation,
                    transformer_type=transformer_type,
                )
                for _ in range(num_layers)
            ]
        )

    def forward(
        self,
        tokens: torch.Tensor,
        cutoff_factors: torch.Tensor,
        use_manual_attention: bool,
    ) -> torch.Tensor:
        for layer in self.layers:
            tokens = layer(tokens, cutoff_factors, use_manual_attention)
        if self.transformer_type == "PreLN":
            tokens = self.final_norm(tokens)
        return tokens


class CartesianTransformer(torch.nn.Module):
    def __init__(
        self,
        hypers,
        d_model: int,
        n_head: int,
        dim_feedforward: int,
        n_layers: int,
        dropout: float,
        n_atomic_species: int,
        is_first,
    ):
        super(CartesianTransformer, self).__init__()
        self.is_first = is_first
        self.cutoff = float(hypers["cutoff"])
        self.cutoff_width = float(hypers["cutoff_width"])
        self.trans = Transformer(
            d_model=d_model,
            num_layers=n_layers,
            n_heads=n_head,
            dim_feedforward=dim_feedforward,
            dropout=dropout,
            activation=torch.nn.SiLU(),
            transformer_type="PostLN",
        )

        self.edge_embedder = nn.Linear(4, d_model)

        if not is_first:
            n_merge = 3
        else:
            n_merge = 2

        self.compress = nn.Sequential(
            nn.Linear(n_merge * d_model, d_model),
            torch.nn.SiLU(),
            nn.Linear(d_model, d_model),
        )

        self.neighbor_embedder = DummyModule()  # for torchscript
        if not is_first:
            self.neighbor_embedder = nn.Embedding(n_atomic_species + 1, d_model)

        self.node_embedder = nn.Embedding(n_atomic_species + 1, d_model)

    def forward(
        self,
        input_messages: torch.Tensor,
        element_indices_nodes: torch.Tensor,
        element_indices_neighbors: torch.Tensor,
        edge_vectors: torch.Tensor,
        padding_mask: torch.Tensor,
        edge_distances: torch.Tensor,
        cutoff_factors: torch.Tensor,
        use_manual_attention: bool,
    ):
        node_elements_embedding = self.node_embedder(element_indices_nodes)
        edge_embedding = [edge_vectors, edge_distances[:, :, None]]
        edge_embedding = torch.cat(edge_embedding, dim=2)
        edge_embedding = self.edge_embedder(edge_embedding)

        if not self.is_first:
            neighbor_elements_embedding = self.neighbor_embedder(
                element_indices_neighbors
            )
            tokens = torch.cat(
                [edge_embedding, neighbor_elements_embedding, input_messages], dim=2
            )
        else:
            neighbor_elements_embedding = torch.empty(
                0, device=edge_vectors.device, dtype=edge_vectors.dtype
            )  # for torch script
            tokens = torch.cat([edge_embedding, input_messages], dim=2)

        tokens = self.compress(tokens)
        tokens = torch.cat([node_elements_embedding[:, None, :], tokens], dim=1)

        padding_mask_with_central_token = torch.ones(
            padding_mask.shape[0], dtype=torch.bool, device=padding_mask.device
        )
        total_padding_mask = torch.cat(
            [padding_mask_with_central_token[:, None], padding_mask], dim=1
        )

        cutoff_subfactors = torch.ones(
            padding_mask.shape[0], device=padding_mask.device
        )
        cutoff_factors = torch.cat([cutoff_subfactors[:, None], cutoff_factors], dim=1)
        cutoff_factors[~total_padding_mask] = 0.0

        cutoff_factors = cutoff_factors[:, None, :]
        cutoff_factors = cutoff_factors.repeat(1, cutoff_factors.shape[2], 1)

        initial_num_tokens = edge_vectors.shape[1]
        max_num_tokens = input_messages.shape[1]

        output_messages = self.trans(
            tokens[:, : (max_num_tokens + 1), :],
            cutoff_factors=cutoff_factors[
                :, : (max_num_tokens + 1), : (max_num_tokens + 1)
            ],
            use_manual_attention=use_manual_attention,
        )
        # print("[NATIVEPET] output_messages", output_messages[0][:6])
        if max_num_tokens < initial_num_tokens:
            padding = torch.zeros(
                output_messages.shape[0],
                initial_num_tokens - max_num_tokens,
                output_messages.shape[2],
                device=output_messages.device,
            )
            output_messages = torch.cat([output_messages, padding], dim=1)

        output_node_embeddings = output_messages[:, 0, :]
        output_edge_embeddings = output_messages[:, 1:, :]
        return output_node_embeddings, output_edge_embeddings


def manual_attention(q, k, v, attn_mask):
    # needed for double backward (training with conservative forces)
    attention_weights = (
        torch.matmul(q, k.transpose(-2, -1)) / (k.size(-1) ** 0.5)
    ) + attn_mask
    attention_weights = attention_weights.softmax(dim=-1)
    attention_output = torch.matmul(attention_weights, v)
    return attention_output



import copy
import logging
import os
import random

import numpy as np
import torch
from scipy.spatial.transform import Rotation
from scipy.spatial.transform import Rotation as R
from scipy.special import roots_legendre
from torch.optim.lr_scheduler import LambdaLR
from torch_geometric.loader import DataListLoader, DataLoader, DynamicBatchSampler


def get_activation(hypers):
    if hypers.ACTIVATION == "mish":
        return torch.nn.Mish()
    if hypers.ACTIVATION == "silu":
        return torch.nn.SiLU()
    raise ValueError("unknown activation")


def log_epoch_stats(epoch, total_epochs, epoch_stats, learning_rate, energies_key):
    """
    Logs the detailed training and validation statistics at the end of each epoch.

    Parameters are the same as in the previous version, with added validation metrics.
    """

    if "forces" in epoch_stats.keys():
        write_forces_error = True
    else:
        write_forces_error = False
    epoch_stats_log = f"Epoch: {epoch} | LR: {learning_rate:.3e} | "

    train_energies_mae = epoch_stats[energies_key]["train"]["mae"]
    train_energies_rmse = epoch_stats[energies_key]["train"]["rmse"]
    if write_forces_error:
        train_forces_mae = epoch_stats["forces"]["train"]["mae"]
        train_forces_rmse = epoch_stats["forces"]["train"]["rmse"]

    val_energies_mae = epoch_stats[energies_key]["val"]["mae"]
    val_energies_rmse = epoch_stats[energies_key]["val"]["rmse"]
    if write_forces_error:
        val_forces_mae = epoch_stats["forces"]["val"]["mae"]
        val_forces_rmse = epoch_stats["forces"]["val"]["rmse"]

    if energies_key == "energies per structure":
        energies_log_key = ""
    elif energies_key == "energies per atom":
        energies_log_key = "/at."
    else:
        energies_log_key = ""

    # epoch_time = epoch_stats["epoch_time"]
    total_time = epoch_stats["elapsed_time"]
    estimated_remaining_time = epoch_stats["estimated_remaining_time"]

    epoch_stats_log += f"V-E-MAE{energies_log_key} {val_energies_mae:.2e} | "
    epoch_stats_log += f"V-E-RMSE{energies_log_key} {val_energies_rmse:.2e} | "
    if write_forces_error:
        epoch_stats_log += f"V-F-MAE {val_forces_mae:.2e} | "
        epoch_stats_log += f"V-F-RMSE {val_forces_rmse:.2e} | "
    epoch_stats_log += f"T-E-MAE{energies_log_key} {train_energies_mae:.2e} | "
    epoch_stats_log += f"T-E-RMSE{energies_log_key} {train_energies_rmse:.2e} | "
    if write_forces_error:
        epoch_stats_log += f"T-F-MAE {train_forces_mae:.2e} | "
        epoch_stats_log += f"T-F-RMSE {train_forces_rmse:.2e} | "

    epoch_stats_log += f"Time/ETA: {total_time:.2f}/{estimated_remaining_time:.2f} s"
    logging.info(epoch_stats_log)


def get_calc_names(all_completed_calcs, current_name):
    name_to_load = None
    name_of_calculation = current_name
    if name_of_calculation in all_completed_calcs:
        name_to_load = name_of_calculation
        for i in range(100000):
            name_now = name_of_calculation + f"_continuation_{i}"
            if name_now not in all_completed_calcs:
                name_to_save = name_now
                break
            name_to_load = name_now
        name_of_calculation = name_to_save
    return name_to_load, name_of_calculation


def set_reproducibility(random_seed, cuda_deterministic):
    torch.manual_seed(random_seed)
    np.random.seed(random_seed)
    random.seed(random_seed)
    os.environ["PYTHONHASHSEED"] = str(random_seed)
    if torch.cuda.is_available():
        torch.cuda.manual_seed(random_seed)
        torch.cuda.manual_seed_all(random_seed)

    if cuda_deterministic and torch.cuda.is_available():
        torch.use_deterministic_algorithms(True)
        torch.backends.cudnn.benchmark = False
        os.environ["CUBLAS_WORKSPACE_CONFIG"] = ":4096:8"


def get_length(delta):
    return np.sqrt(np.sum(delta * delta))


class ModelKeeper:
    def __init__(self):
        self.best_model = None
        self.best_error = None
        self.best_epoch = None
        self.additional_info = None

    def update(self, model_now, error_now, epoch_now, additional_info=None):
        if (self.best_error is None) or (error_now < self.best_error):
            self.best_error = error_now
            original_device = next(model_now.parameters()).device
            model_now.to("cpu")
            self.best_model = copy.deepcopy(model_now)
            model_now.to(original_device)
            self.best_epoch = epoch_now
            self.additional_info = additional_info


class Accumulator:
    def __init__(self):
        self.values = None

    def update(self, values_now):
        if isinstance(values_now, torch.Tensor):
            values_now = [values_now]

        if self.values is None:
            self.values = [[] for _ in range(len(values_now))]

        for index, value_now in enumerate(values_now):
            if isinstance(value_now, torch.Tensor):
                value_now = value_now.data.cpu().to(torch.float32).numpy()
            self.values[index].append(value_now)

    def consist_of_nones(self, el):
        has_none = False
        has_value = False
        for index in range(len(el)):
            if el[index] is None:
                has_none = True
            else:
                has_value = True
        if has_none and has_value:
            raise ValueError("Some values are None, some are not")
        return has_none

    def flush(self):
        result = []
        for el in self.values:
            if self.consist_of_nones(el):
                result.append(None)
            else:
                result.append(np.concatenate(el, axis=0))
        self.values = None
        return result


class Logger:
    def __init__(self, support_missing_values, use_shift_agnostic_loss, device):
        self.predictions = []
        self.targets = []
        self.support_missing_values = support_missing_values
        self.use_shift_agnostic_loss = use_shift_agnostic_loss
        self.device = device

    def update(self, predictions_now, targets_now):
        if isinstance(predictions_now, dict):
            predictions_now = predictions_now["prediction"]
        self.predictions.append(predictions_now.data.cpu().to(torch.float32).numpy())
        self.targets.append(targets_now.data.cpu().to(torch.float32).numpy())

    def flush(self):
        if not self.use_shift_agnostic_loss:
            self.predictions = np.concatenate(self.predictions, axis=0)
            self.targets = np.concatenate(self.targets, axis=0)

            output = {}
            output["rmse"] = get_rmse(
                self.predictions,
                self.targets,
                support_missing_values=self.support_missing_values,
            )
            output["mae"] = get_mae(
                self.predictions,
                self.targets,
                support_missing_values=self.support_missing_values,
            )
            output["relative rmse"] = get_relative_rmse(
                self.predictions,
                self.targets,
                support_missing_values=self.support_missing_values,
            )

            self.predictions = []
            self.targets = []
            return output
        else:
            if self.support_missing_values:
                raise ValueError(
                    "Shift agnostic loss is not supported with missing values"
                )

            total_mse = 0.0
            total_mae = 0.0
            num_samples = 0

            with torch.no_grad():
                for prediction_batch, target_batch in zip(
                    self.predictions, self.targets
                ):
                    prediction_batch = torch.tensor(
                        prediction_batch,
                        device=self.device,
                        dtype=torch.get_default_dtype(),
                    )
                    target_batch = torch.tensor(
                        target_batch,
                        device=self.device,
                        dtype=torch.get_default_dtype(),
                    )
                    batch_size = len(target_batch)
                    current_mse = (
                        get_shift_agnostic_mse(prediction_batch, target_batch)
                        * batch_size
                    )
                    current_mae = (
                        get_shift_agnostic_mae(prediction_batch, target_batch)
                        * batch_size
                    )
                    current_mse = float(current_mse)
                    current_mae = float(current_mae)
                    total_mse += current_mse
                    total_mae += current_mae
                    num_samples += batch_size

            rmse = np.sqrt(total_mse / num_samples)
            mae = total_mae / num_samples

            output = {}
            output["rmse"] = rmse
            output["mae"] = mae

            self.predictions = []
            self.targets = []
            return output


class FullLogger:
    def __init__(self, support_missing_values, use_shift_agnostic_loss, device):
        self.train_logger = Logger(
            support_missing_values, use_shift_agnostic_loss, device
        )
        self.val_logger = Logger(
            support_missing_values, use_shift_agnostic_loss, device
        )

    def flush(self):
        return {"train": self.train_logger.flush(), "val": self.val_logger.flush()}


def get_rotations(indices, global_aug=False):
    if global_aug:
        num = np.max(indices) + 1
    else:
        num = indices.shape[0]

    rotations = Rotation.random(num).as_matrix()
    rotations[np.random.randn(rotations.shape[0]) >= 0] *= -1

    if global_aug:
        return rotations[indices]
    else:
        return rotations


def get_shift_agnostic_mse(predictions, targets):
    if predictions.shape[1] < targets.shape[1]:
        smaller = predictions
        bigger = targets
    else:
        smaller = targets
        bigger = predictions

    bigger_unfolded = bigger.unfold(1, smaller.shape[1], 1)
    smaller_expanded = smaller[:, None, :]
    delta = smaller_expanded - bigger_unfolded
    losses = torch.mean(delta * delta, dim=2)
    losses, _ = torch.min(losses, dim=1)
    result = torch.mean(losses)
    return result


def get_shift_agnostic_mae(predictions, targets):
    if predictions.shape[1] < targets.shape[1]:
        smaller = predictions
        bigger = targets
    else:
        smaller = targets
        bigger = predictions

    bigger_unfolded = bigger.unfold(1, smaller.shape[1], 1)
    smaller_expanded = smaller[:, None, :]
    delta = smaller_expanded - bigger_unfolded
    losses = torch.mean(torch.abs(delta), dim=2)
    losses, _ = torch.min(losses, dim=1)
    result = torch.mean(losses)
    return result


def get_loss(predictions, targets, support_missing_values, use_shift_agnostic_loss):
    if use_shift_agnostic_loss:
        if support_missing_values:
            raise NotImplementedError(
                "shift agnostic loss is not yet supported with missing values"
            )
        else:
            return get_shift_agnostic_mse(predictions, targets)
    else:
        if isinstance(predictions, dict):
            predictions = predictions["prediction"]
        if support_missing_values:
            delta = predictions - targets
            mask_nan = torch.isnan(targets)
            delta[mask_nan] = 0.0
            mask_not_nan = torch.logical_not(mask_nan)
            return torch.sum(delta * delta) / torch.sum(mask_not_nan)
        else:
            delta = predictions - targets
            return torch.mean(delta * delta)


def get_rmse(predictions, targets, support_missing_values=False):
    if support_missing_values:
        delta = predictions - targets
        mask_nan = np.isnan(targets)
        delta[mask_nan] = 0.0
        mask_not_nan = np.logical_not(mask_nan)
        return np.sqrt(np.sum(delta * delta) / np.sum(mask_not_nan))
    else:
        delta = predictions - targets
        return np.sqrt(np.mean(delta * delta))


def get_mae(predictions, targets, support_missing_values=False):
    if support_missing_values:
        delta = predictions - targets
        mask_nan = np.isnan(targets)
        delta[mask_nan] = 0.0
        mask_not_nan = np.logical_not(mask_nan)
        return np.sum(np.abs(delta)) / np.sum(mask_not_nan)
    else:
        delta = predictions - targets
        return np.mean(np.abs(delta))


def get_relative_rmse(predictions, targets, support_missing_values=False):
    rmse = get_rmse(predictions, targets, support_missing_values=support_missing_values)
    return rmse / get_rmse(
        np.mean(targets), targets, support_missing_values=support_missing_values
    )


def get_scheduler(optim, FITTING_SCHEME):
    def func_lr_scheduler(epoch):
        if epoch < FITTING_SCHEME.EPOCHS_WARMUP:
            return epoch / FITTING_SCHEME.EPOCHS_WARMUP
        delta = epoch - FITTING_SCHEME.EPOCHS_WARMUP
        num_blocks = delta // FITTING_SCHEME.SCHEDULER_STEP_SIZE
        return 0.5 ** (num_blocks)

    scheduler = LambdaLR(optim, func_lr_scheduler)
    return scheduler


def load_checkpoint(model, optim, scheduler, checkpoint_path):
    checkpoint = torch.load(checkpoint_path)
    model.load_state_dict(checkpoint["model_state_dict"])
    optim.load_state_dict(checkpoint["optim_state_dict"])
    scheduler.load_state_dict(checkpoint["scheduler_state_dict"])


def get_data_loaders(train_graphs, val_graphs, FITTING_SCHEME):
    def seed_worker(worker_id):
        worker_seed = torch.initial_seed() % 2**32
        np.random.seed(worker_seed)
        random.seed(worker_seed)

    g = torch.Generator()
    g.manual_seed(FITTING_SCHEME.RANDOM_SEED)

    if FITTING_SCHEME.BALANCED_DATA_LOADER:
        train_sampler = DynamicBatchSampler(
            train_graphs,
            max_num=FITTING_SCHEME.ATOMIC_BATCH_SIZE,
            mode="node",
            shuffle=True,
        )
        val_sampler = DynamicBatchSampler(
            val_graphs,
            max_num=FITTING_SCHEME.ATOMIC_BATCH_SIZE,
            mode="node",
            shuffle=False,
        )

        if FITTING_SCHEME.MULTI_GPU:
            train_loader = DataListLoader(
                train_graphs,
                batch_sampler=train_sampler,
                worker_init_fn=seed_worker,
                generator=g,
            )
            val_loader = DataListLoader(
                val_graphs,
                batch_sampler=val_sampler,
                worker_init_fn=seed_worker,
                generator=g,
            )
        else:
            train_loader = DataLoader(
                train_graphs,
                batch_sampler=train_sampler,
                worker_init_fn=seed_worker,
                generator=g,
            )
            val_loader = DataLoader(
                val_graphs,
                batch_sampler=val_sampler,
                worker_init_fn=seed_worker,
                generator=g,
            )
    else:
        if FITTING_SCHEME.MULTI_GPU:
            train_loader = DataListLoader(
                train_graphs,
                batch_size=FITTING_SCHEME.STRUCTURAL_BATCH_SIZE,
                shuffle=True,
                worker_init_fn=seed_worker,
                generator=g,
            )
            val_loader = DataListLoader(
                val_graphs,
                batch_size=FITTING_SCHEME.STRUCTURAL_BATCH_SIZE,
                shuffle=False,
                worker_init_fn=seed_worker,
                generator=g,
            )
        else:
            train_loader = DataLoader(
                train_graphs,
                batch_size=FITTING_SCHEME.STRUCTURAL_BATCH_SIZE,
                shuffle=True,
                worker_init_fn=seed_worker,
                generator=g,
            )
            val_loader = DataLoader(
                val_graphs,
                batch_size=FITTING_SCHEME.STRUCTURAL_BATCH_SIZE,
                shuffle=False,
                worker_init_fn=seed_worker,
                generator=g,
            )

    return train_loader, val_loader


def get_optimizer(model, FITTING_SCHEME):
    if FITTING_SCHEME.USE_WEIGHT_DECAY:
        optim = torch.optim.AdamW(
            model.parameters(),
            lr=FITTING_SCHEME.INITIAL_LR,
            weight_decay=FITTING_SCHEME.WEIGHT_DECAY,
        )
    else:
        optim = torch.optim.Adam(model.parameters(), lr=FITTING_SCHEME.INITIAL_LR)
    return optim


def get_rotational_discrepancy(all_predictions):
    predictions_mean = np.mean(all_predictions, axis=0)
    predictions_discrepancies = all_predictions - predictions_mean[np.newaxis]
    # correction for unbiased estimate
    correction = all_predictions.shape[0] / (all_predictions.shape[0] - 1)
    predictions_std = np.sqrt(np.mean(predictions_discrepancies**2) * correction)

    # biased estimate, kind of a mess with the unbiased one
    predictions_mad = np.mean(np.abs(predictions_discrepancies))
    return predictions_std, predictions_mad


def report_accuracy(
    all_predictions,
    ground_truth,
    target_name,
    verbose,
    specify_per_component,
    target_type,
    n_atoms=None,
    support_missing_values=False,
):
    predictions_mean = np.mean(all_predictions, axis=0)

    if specify_per_component:
        spec = "per component"
    else:
        spec = ""

    if ground_truth is not None:
        mae = get_mae(
            predictions_mean,
            ground_truth,
            support_missing_values=support_missing_values,
        )
        rmse = get_rmse(
            predictions_mean,
            ground_truth,
            support_missing_values=support_missing_values,
        )
        print(f"{target_name} mae {spec}: {mae}")
        print(f"{target_name} rmse {spec}: {rmse}")
    else:
        print(
            f"ground truth target for {target_name} is not provided "
            "(or is provided with a wrong key). Thus, it is impossible "
            "to estimate the error between predictions and ground truth target"
        )

    if all_predictions.shape[0] > 1:
        pred_std, pred_mad = get_rotational_discrepancy(all_predictions)
        if verbose:
            str_mae = "rotational discrepancy mad (aka mae)"
            str_rmse = "rotational discrepancy std (aka rmse)"
            print(f"{target_name} {str_mae} {spec}: {pred_mad}")

            print(f"{target_name} {str_rmse} {spec}: {pred_std} ")

    if target_type == "structural":
        if len(predictions_mean.shape) == 1:
            predictions_mean = predictions_mean[:, np.newaxis]
        predictions_mean_per_atom = predictions_mean / n_atoms[:, np.newaxis]

        if ground_truth is not None:
            if len(ground_truth.shape) == 1:
                ground_truth = ground_truth[:, np.newaxis]

            ground_truth_per_atom = ground_truth / n_atoms[:, np.newaxis]
            mae = get_mae(
                predictions_mean_per_atom,
                ground_truth_per_atom,
                support_missing_values=support_missing_values,
            )
            rmse = get_rmse(
                predictions_mean_per_atom,
                ground_truth_per_atom,
                support_missing_values=support_missing_values,
            )
            print(f"{target_name} mae per atom {spec}: " + f"{mae}")

            print(f"{target_name} rmse per atom {spec}: {rmse}")

        if all_predictions.shape[0] > 1:
            if len(all_predictions.shape) == 2:
                all_predictions = all_predictions[:, :, np.newaxis]
            all_predictions_per_atom = (
                all_predictions / n_atoms[np.newaxis, :, np.newaxis]
            )
            predictions_std_per_atom, predictions_mad_per_atom = (
                get_rotational_discrepancy(all_predictions_per_atom)
            )
            if verbose:
                str_mae = "rotational discrepancy mad (aka mae)"
                str_rmse = "rotational discrepancy std (aka rmse)"
                print(f"{target_name} {str_mae} {spec}: {predictions_mad_per_atom}")
                print(f"{target_name} {str_rmse} {spec}: {predictions_std_per_atom} ")


class NeverRun(torch.nn.Module):
    """Dummy torch module to make torchscript happy.
    This model should never be run"""

    def __init__(self):
        super(NeverRun, self).__init__()

    def forward(self, x) -> torch.Tensor:
        raise RuntimeError("This model should never be run")


def get_quadrature(L):
    matrices, weights = [], []
    for theta_index in range(0, 2 * L - 1):
        for w_index in range(0, 2 * L - 1):
            theta = 2 * np.pi * theta_index / (2 * L - 1)
            w = 2 * np.pi * w_index / (2 * L - 1)
            roots_legendre_now, weights_now = roots_legendre(L)
            all_v = np.arccos(roots_legendre_now)
            for v, weight in zip(all_v, weights_now):
                weights.append(weight)
                angles = [theta, v, w]
                rotation = R.from_euler("xyz", angles, degrees=False)
                rotation_matrix = rotation.as_matrix()
                matrices.append(rotation_matrix)

    return matrices, weights


def dtype2string(dtype):
    if dtype == torch.float32:
        return "float32"
    if dtype == torch.float16:
        return "float16"
    if dtype == torch.bfloat16:
        return "bfloat16"

    raise ValueError("unknown dtype")


def string2dtype(string):
    if string == "float32":
        return torch.float32
    if string == "float16":
        return torch.float16
    if string == "bfloat16":
        return torch.bfloat16

    raise ValueError("unknown dtype")


def get_quadrature_predictions(batch, model, quadrature_order, dtype):
    x_initial = batch.x.clone()
    all_energies, all_forces = [], []
    rotations, weights = get_quadrature(quadrature_order)
    for rotation in rotations:
        rotation = torch.tensor(rotation, device=batch.x.device, dtype=dtype)
        batch_rotations = rotation[None, :].repeat(batch.num_nodes, 1, 1)
        batch.x = torch.bmm(x_initial, batch_rotations)
        prediction_energy, prediction_forces = model(
            batch, augmentation=False, create_graph=False
        )
        all_energies.append(prediction_energy.data.cpu().numpy())
        all_forces.append(prediction_forces.data.cpu().numpy())

    energy_mean, forces_mean, total_weight = 0.0, 0.0, 0.0
    for energy, forces, weight in zip(all_energies, all_forces, weights):
        energy_mean += energy * weight
        forces_mean += forces * weight
        total_weight += weight
    energy_mean /= total_weight
    forces_mean /= total_weight
    return energy_mean, forces_mean



import json
from pathlib import Path

import pytest
import torch
from jsonschema.exceptions import ValidationError
from metatensor.torch import Labels
from metatensor.torch.atomistic import (
    MetatensorAtomisticModel,
    ModelCapabilities,
    ModelEvaluationOptions,
    ModelMetadata,
    ModelOutput,
    System,
)

from metatrain.pet import PET as WrappedPET
from metatrain.pet.modules.hypers import Hypers
from metatrain.pet.modules.pet import PET
from metatrain.utils.architectures import get_default_hypers
from metatrain.utils.data import DatasetInfo
from metatrain.utils.data.target_info import (
    get_energy_target_info,
    get_generic_target_info,
)
from metatrain.utils.jsonschema import validate
from metatrain.utils.neighbor_lists import (
    get_requested_neighbor_lists,
    get_system_with_neighbor_lists,
)


DEFAULT_HYPERS = get_default_hypers("pet")
with open(Path(__file__).parents[1] / "schema-hypers.json", "r") as f:
    SCHEMA_HYPERS = json.load(f)


@pytest.mark.parametrize(
    "new_option",
    [
        "ATOMIC_BATCH_SIZE",
        "EPOCH_NUM_ATOMIC",
        "SCHEDULER_STEP_SIZE_ATOMIC",
        "EPOCHS_WARMUP_ATOMIC",
    ],
)
def test_exclusive_hypers(new_option):
    """Test that the `_ATOMIC` is mutually exclusive."""

    options = {
        "training": {
            "SCHEDULER_STEP_SIZE": 1,
            "EPOCH_NUM": 1,
            "STRUCTURAL_BATCH_SIZE": 1,
            "EPOCHS_WARMUP": 1,
        }
    }

    validate(instance=options, schema=SCHEMA_HYPERS)

    options["training"][new_option] = 1
    with pytest.raises(ValidationError, match="should not be valid under"):
        validate(instance=options, schema=SCHEMA_HYPERS)


def test_prediction():
    """Tests that the model runs without errors."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = WrappedPET(DEFAULT_HYPERS["model"], dataset_info)
    ARCHITECTURAL_HYPERS = Hypers(model.hypers)
    raw_pet = PET(ARCHITECTURAL_HYPERS, 0.0, len(model.atomic_types))
    model.set_trained_model(raw_pet)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    requested_neighbor_lists = get_requested_neighbor_lists(model)
    system = get_system_with_neighbor_lists(system, requested_neighbor_lists)

    evaluation_options = ModelEvaluationOptions(
        length_unit=dataset_info.length_unit,
        outputs={"energy": ModelOutput()},
    )

    capabilities = ModelCapabilities(
        length_unit="Angstrom",
        atomic_types=model.atomic_types,
        outputs={
            "energy": ModelOutput(
                quantity="energy",
                unit="eV",
            )
        },
        interaction_range=DEFAULT_HYPERS["model"]["N_GNN_LAYERS"]
        * DEFAULT_HYPERS["model"]["R_CUT"],
        dtype="float32",
        supported_devices=["cpu", "cuda"],
    )

    model = MetatensorAtomisticModel(model.eval(), ModelMetadata(), capabilities)
    model(
        [system],
        evaluation_options,
        check_consistency=True,
    )


def test_per_atom_predictions_functionality():
    """Tests that the model can do predictions in
    per-atom mode."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = WrappedPET(DEFAULT_HYPERS["model"], dataset_info)
    ARCHITECTURAL_HYPERS = Hypers(model.hypers)
    raw_pet = PET(ARCHITECTURAL_HYPERS, 0.0, len(model.atomic_types))
    model.set_trained_model(raw_pet)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    requested_neighbor_lists = get_requested_neighbor_lists(model)
    system = get_system_with_neighbor_lists(system, requested_neighbor_lists)

    evaluation_options = ModelEvaluationOptions(
        length_unit=dataset_info.length_unit,
        outputs={"energy": ModelOutput()},
    )

    capabilities = ModelCapabilities(
        length_unit="Angstrom",
        atomic_types=model.atomic_types,
        outputs={
            "energy": ModelOutput(
                quantity="energy",
                unit="eV",
                per_atom=True,
            )
        },
        interaction_range=DEFAULT_HYPERS["model"]["N_GNN_LAYERS"]
        * DEFAULT_HYPERS["model"]["R_CUT"],
        dtype="float32",
        supported_devices=["cpu", "cuda"],
    )

    model = MetatensorAtomisticModel(model.eval(), ModelMetadata(), capabilities)
    model(
        [system],
        evaluation_options,
        check_consistency=True,
    )


def test_selected_atoms_functionality():
    """Tests that the model can do predictions for a selected
    subset of the atoms in the system."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )
    model = WrappedPET(DEFAULT_HYPERS["model"], dataset_info)
    ARCHITECTURAL_HYPERS = Hypers(model.hypers)
    raw_pet = PET(ARCHITECTURAL_HYPERS, 0.0, len(model.atomic_types))
    model.set_trained_model(raw_pet)

    system = System(
        types=torch.tensor([6, 6]),
        positions=torch.tensor([[0.0, 0.0, 0.0], [0.0, 0.0, 1.0]]),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )
    requested_neighbor_lists = get_requested_neighbor_lists(model)
    system = get_system_with_neighbor_lists(system, requested_neighbor_lists)

    evaluation_options = ModelEvaluationOptions(
        length_unit=dataset_info.length_unit,
        outputs={"energy": ModelOutput()},
    )

    capabilities = ModelCapabilities(
        length_unit="Angstrom",
        atomic_types=model.atomic_types,
        outputs={
            "energy": ModelOutput(
                quantity="energy",
                unit="eV",
            )
        },
        interaction_range=DEFAULT_HYPERS["model"]["N_GNN_LAYERS"]
        * DEFAULT_HYPERS["model"]["R_CUT"],
        dtype="float32",
        supported_devices=["cpu", "cuda"],
    )

    selected_atoms = Labels(
        ["system", "atom"],
        torch.tensor([[0, a] for a in range(len(system)) if a % 2 == 0]),
    )

    evaluation_options = ModelEvaluationOptions(
        length_unit=capabilities.length_unit,
        outputs=capabilities.outputs,
        selected_atoms=selected_atoms,
    )

    model = MetatensorAtomisticModel(model.eval(), ModelMetadata(), capabilities)
    model(
        [system],
        evaluation_options,
        check_consistency=True,
    )


@pytest.mark.parametrize("per_atom", [True, False])
def test_vector_output(per_atom):
    """Tests that the model can predict a (spherical) vector output."""

    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={
            "forces": get_generic_target_info(
                {
                    "quantity": "forces",
                    "unit": "",
                    "type": {
                        "spherical": {"irreps": [{"o3_lambda": 1, "o3_sigma": 1}]}
                    },
                    "num_subtargets": 100,
                    "per_atom": per_atom,
                }
            )
        },
    )

    with pytest.raises(ValueError, match="PET only supports total-energy-like outputs"):
        WrappedPET(DEFAULT_HYPERS["model"], dataset_info)


def test_output_features():
    """Tests that the model can output its features and last-layer features."""
    dataset_info = DatasetInfo(
        length_unit="Angstrom",
        atomic_types=[1, 6, 7, 8],
        targets={"energy": get_energy_target_info({"unit": "eV"})},
    )

    model = WrappedPET(DEFAULT_HYPERS["model"], dataset_info)
    ARCHITECTURAL_HYPERS = Hypers(model.hypers)
    raw_pet = PET(ARCHITECTURAL_HYPERS, 0.0, len(model.atomic_types))
    model.set_trained_model(raw_pet)

    system = System(
        types=torch.tensor([6, 1, 8, 7]),
        positions=torch.tensor(
            [[0.0, 0.0, 0.0], [0.0, 0.0, 1.0], [0.0, 0.0, 2.0], [0.0, 0.0, 3.0]],
        ),
        cell=torch.zeros(3, 3),
        pbc=torch.tensor([False, False, False]),
    )

    requested_neighbor_lists = get_requested_neighbor_lists(model)
    system = get_system_with_neighbor_lists(system, requested_neighbor_lists)

    # last-layer features per atom:
    ll_output_options = ModelOutput(
        quantity="",
        unit="unitless",
        per_atom=True,
    )
    outputs = model(
        [system],
        {
            "energy": ModelOutput(quantity="energy", unit="eV", per_atom=True),
            "mtt::aux::energy_last_layer_features": ll_output_options,
            "features": ll_output_options,
        },
    )
    assert "energy" in outputs
    assert "mtt::aux::energy_last_layer_features" in outputs
    assert "features" in outputs
    last_layer_features = outputs["mtt::aux::energy_last_layer_features"].block()
    features = outputs["features"].block()
    assert last_layer_features.samples.names == ["system", "atom"]
    assert last_layer_features.values.shape == (
        4,
        768,  # 768 = 3 (gnn layers) * 256 (128 for edge repr, 128 for node repr)
    )
    assert last_layer_features.properties.names == ["properties"]
    assert features.samples.names == ["system", "atom"]
    assert features.values.shape == (
        4,
        768,  # 768 = 3 (gnn layers) * 256 (128 for edge repr, 128 for node repr)
    )
    assert features.properties.names == ["properties"]

    # last-layer features per system:
    ll_output_options = ModelOutput(
        quantity="",
        unit="unitless",
        per_atom=False,
    )
    outputs = model(
        [system],
        {
            "energy": ModelOutput(quantity="energy", unit="eV", per_atom=True),
            "mtt::aux::energy_last_layer_features": ll_output_options,
            "features": ll_output_options,
        },
    )
    assert "energy" in outputs
    assert "mtt::aux::energy_last_layer_features" in outputs
    assert "features" in outputs
    assert outputs["mtt::aux::energy_last_layer_features"].block().samples.names == [
        "system"
    ]
    assert outputs["mtt::aux::energy_last_layer_features"].block().values.shape == (
        1,
        768,  # 768 = 3 (gnn layers) * 256 (128 for edge repr, 128 for node repr)
    )
    assert outputs["mtt::aux::energy_last_layer_features"].block().properties.names == [
        "properties",
    ]
    assert outputs["features"].block().samples.names == ["system"]
    assert outputs["features"].block().values.shape == (
        1,
        768,  # 768 = 3 (gnn layers) * 256 (128 for edge repr, 128 for node repr)
    )
    assert outputs["features"].block().properties.names == ["properties"]
